/*
 * @(#)ExtendedList.java    19.6.1999
 */


package corewar;

import java.awt.List;
import java.awt.Font;
import java.awt.event.ItemEvent;
import java.util.Vector;

import java.awt.Frame;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ItemListener;

/**
 * Die abstrakte Klasse ExtendedList ist eine visuelle Liste, deren Elemente
 * tabellenfrmig angezeigt werden. Die Klasse bernimmt auch die Verwaltung
 * der Objekte in einem Vector. Ableitende Klassen mssen die abstrakten
 * Funktionen getColumnWidths(), getTitles() und getEntries(Object)
 * berschreiben. 
 *
 * @author  Ivo Kasiuk
 */
public abstract class ExtendedList extends List {

    /* Anzahl der fr die berschriften verwendeten Zeilen. */
    private static final int HEADER_LINES = 2;
    
    private Vector elements;
    
    /**
     * Erstellt eine neue Liste.
     * @param lines  Die sichtbare Hhe der Liste.
     */
    public ExtendedList(int lines) {
        super(lines);
        elements = new Vector();
        setFont(Font.decode("Monospaced"));
        super.add(formatLine(getTitles()));
        super.add(makeLine());
    }

    /*
     * Erzeugt eine Trennzeile zur Trennung zwischen Titelzeile und Eintrgen.
     */
    private String makeLine() {
        String line   = "";
        int[]  widths = getColumnWidths();
        for (int i=0; i<widths.length; i++) {
            for (int j=0; j<widths[i]; j++) {
                line += '=';
            }
            line += ' ';
        }
        return line;
    }

    /*
     * Bildet aus den Zeileneintrgen entries eine Tabellenzeile.
     */
    private String formatLine(String[] entries) {
        String result = "";
        int[]  widths = getColumnWidths();
        for (int i=0; i<entries.length; i++) {
            if (entries[i].length() >= widths[i]) {
                result += entries[i].substring(0, widths[i]);
            } else {
                result += entries[i];
                for (int j=0; j<(widths[i] - entries[i].length()); j++) {
                    result += ' ';
                }
            }
            result += ' ';
        }
        return result;
    }
    
    /**
     * Liefert den Vector von Elementen fr diese Liste.
     */
    public Vector getElements() {
        return elements;
    }

    /**
     * bernimmt einen Vector von Elementen fr diese Liste.
     * @param elements  Der Vector mit den Listenelementen.
     */
    public void setElements(Vector elements) {
        this.elements = elements;
        super.removeAll();
        super.add(formatLine(getTitles()));
        super.add(makeLine());
        for (int i=0; i<elements.size(); i++) {
            super.add(formatLine(getEntries(elements.elementAt(i))));
        }
    }

    /**
     * Fgt ein Element am Ende der Liste hinzu.
     * @param element  Das neue Listenelement.
     */
    public void addElement(Object element) {
        elements.addElement(element);
        super.add(formatLine(getEntries(element)));
    }

    /**
     * Fgt ein Element in die Liste ein.
     * @param element  Das neue Listenelement.
     * @param index    Die Position, an der das Element eingefgt werden soll.
     * @exception IndexOutOfBoundsException  wenn index einen ungltigen Wert
     *                hat.
     */
    public void addElementAt(Object element, int index)
            throws IndexOutOfBoundsException {
        elements.insertElementAt(element, index);
        super.add(formatLine(getEntries(element)), index + HEADER_LINES);
    }

    /**
     * Entfernt ein Element aus der Liste.
     * @param index  Die Position des zu entfernenden Elements in der Liste.
     * @exception IndexOutOfBoundsException  Wenn index einen ungltigen Wert
     *                hat.
     */
    public void removeElementAt(int index) throws IndexOutOfBoundsException {
        elements.removeElementAt(index);
        super.remove(index + HEADER_LINES);
    }

    /**
     * Lscht alle Elemente aus der Liste.
     */
    public void removeAllElements() {
        elements.removeAllElements();
        super.removeAll();
        super.add(formatLine(getTitles()));
        super.add(makeLine());
    }

    /**
     * Aktualisiert die Bildschirmdarstellung fr eine Zeile.
     * @param index  Die Nummer der zu aktualisierenden Zeile.
     * @exception IndexOutOfBoundsException  Wenn index einen ungltigen Wert
     *                hat.
     */
    public void refreshLine(int index) throws IndexOutOfBoundsException {
        replaceItem(formatLine(getEntries(elements.elementAt(index))),
                    index + HEADER_LINES);
    }

    /**
     * Aktualisiert die Bildschirmdarstellung fr die ganze Liste.
     */
    public void refreshAll() {
        replaceItem(formatLine(getTitles()), 0);
        replaceItem(makeLine(), 1);
        for(int i=0; i<elements.size(); i++) {
            refreshLine(i);
        }
    }

    /**
     * Liefert die Spaltenbreiten der Tabelle.
     * @return  Ein Array von Spaltenbreiten int Zeichen.
     */
    protected abstract int[] getColumnWidths();

    /**
     * Liefert die berschriften der Spalten.
     * @return  Die berschriften fr die Spalten als Array.
     */
    protected abstract String[] getTitles();

    /**
     * Liefert die Eintrge fr eine Zeile.
     * @param element  Das Objekt, das diese Zeile reprsentiert.
     * @return  Die Eintrge fr diese Zeile als Array.
     */
    protected abstract String[] getEntries(Object element);

    /**
     * Verschiebt die Zeilennummer so, da die erste Spalte der Tabelle die
     * Zeilennummer 0 hat. Die Titelzeilen haben also negative Zeilennummern.
     */
    public int getSelectedIndex() {
        return super.getSelectedIndex() - HEADER_LINES;
    }
    
    /**
     * Verschiebt die Zeilennummer so, da die erste Spalte der Tabelle die
     * Zeilennummer 0 hat. Nachrichten, die sich auf die Titelzeilen beziehen,
     * knnen also daran erkannt werden, da sie eine negative Zeilennummer
     * haben.
     */
    protected void processItemEvent(ItemEvent event) {
        int       item     = ((Integer) event.getItem()).intValue();
        ItemEvent newEvent = new ItemEvent(event.getItemSelectable(),
                ItemEvent.ITEM_STATE_CHANGED,
                new Integer(item - HEADER_LINES), event.getStateChange());
        super.processItemEvent(newEvent);
    }

    /*    
    public static void main(String[] args) {
        ExtendedList list = new ExtendedList(5) {
            protected int[] getColumnWidths() {
                int[] result = { 10, 5 };
                return result;
            }
            protected String[] getTitles() {
                String[] result = { "Das ist die erste Spalte", "zwei" };
                return result;
            }
            protected String[] getEntries(Object element) {
                String[] result = { "Hallo!", "Viel zu lang" };
                return result;
            }
        };
        Frame f = new Frame("Test");
        f.add(list);
        list.addElement(new Object());
        list.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent event) {
                System.out.println(event);
            }
        });
        f.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                System.exit(0);
            }
        });
        f.pack();
        f.setVisible(true);
    }
    */
}
