/*
 * @(#)NetworkSocket.java    18.6.1999
 */


package corewar;

import java.io.OutputStream;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.net.Socket;
import java.net.InetAddress;
import java.net.ConnectException;

/**
 * Die Klasse NetworkSocket stellt eine Netzwerkverbindung dar und ermglicht
 * das stndige Warten auf Nachrichten mit Hilfe eines eigenen Threads.
 * Abgeleitete Klassen sollten handleMessage(byte) berschreiben, um auf
 * ankommende Nachrichten zu reagieren.
 *
 * @author  Ivo Kasiuk
 */
public class NetworkSocket extends Thread implements CWProtocol {

    /* Der Socket fr diese Netzwerkverbindung. */
    private Socket socket;

    /* Der ObjectOutputStream fr diese Netzwerkverbindung. */
    private ObjectOutputStream outStream;

    /* Der ObjectInputStream fr diese Netzwerkverbindung. */
    private ObjectInputStream  inStream;

    /* Signalisiert, da die Verbindung beendet wird. */
    protected boolean aborted = false;

    /**
     * Einen neuen Socket einrichten und Startkommunikation durchfhren.
     * @param address  Die Adresse des Servers.
     * @param port     Die Portnr. des Servers.
     * @param id       Bytes, die zur Identifikation an den Server gesendet
     *                 werden sollen.
     * @exception IOException  Wenn whrend der Startkommunikation ein Fehler
     *                auftritt.
     */
    public NetworkSocket(InetAddress address, int port, byte[] id)
            throws IOException {
        this(address, port);
        initCommunication(id);
    }

    /**
     * Einen neuen Socket einrichten und Startkommunikation durchfhren.
     * @param name     Der Hostname des Servers.
     * @param port     Die Portnr. des Servers.
     * @param id       Bytes, die zur Identifikation an den Server gesendet
     *                 werden sollen.
     * @exception IOException  Wenn whrend der Startkommunikation ein Fehler
     *                auftritt.
     */
    public NetworkSocket(String name, int port, byte[] id) throws IOException {
        socket = new Socket(name, port);
        initCommunication(id);
    }
    
    /**
     * Richtet einen neuen Socket ein. Es wird keine Startkommunikation
     * durchgefhrt.
     * @param address  Die Adresse des Servers.
     * @param port     Die Portnr. des Servers.
     */
    public NetworkSocket(InetAddress address, int port) throws IOException {
        socket = new Socket(address, port);
    }

    /**
     * Einen bestehenden Socket bernehmen, der von einem ServerSocket per
     * accept() erzeugt wurde und Startkommunikation durchfhren.
     * @param client    Der Socket, der die Verbindung zum Client darstellt.
     * @param wantedID  Die erwartete Identifikation.
     * @exception IOException  Wenn whrend der Startkommunikation ein Fehler
     *                auftritt oder die Identifikation des Clients ungltig ist.
     */
    public NetworkSocket(Socket client, byte[] wantedID) throws IOException {
        OutputStream out = client.getOutputStream();
        InputStream  in  = client.getInputStream();
        byte[]       id  = new byte[wantedID.length];
        
        socket = client;
        client.setSoTimeout(CONNECT_TIMEOUT);
        in.read(id);
        for (int i=0; i<wantedID.length; i++) {
            if (id[i] != wantedID[i]) {
                out.write(BYE);
                client.close();
                throw new ConnectException("Ungltige Anmeldung");
            }
        }
        out.write(CONNECT_OK);
        outStream = new ObjectOutputStream(out);
        outStream.flush();
        inStream  = new ObjectInputStream(in);
        client.setSoTimeout(0);
    }

    /**
     * Einen bestehenden Socket bernehmen, dessen Identifikation bereits
     * gelesen und verifiziert ist. Es wird lediglich CONNECT_OK gesendet und
     * der ObjectOutputStream und ObjectInputStream eingerichtet.
     * @param client  Der Socket, der eine Verbindung zum Client darstellt.
     * @exception IOException  wenn whrend der Kommunikation ein Fehler
     *                auftritt.
     */
    public NetworkSocket(Socket client) throws IOException {
        OutputStream out = client.getOutputStream();

        socket = client;
        client.setSoTimeout(CONNECT_TIMEOUT);
        out.write(CONNECT_OK);
        outStream = new ObjectOutputStream(out);
        outStream.flush();
        inStream  = new ObjectInputStream(client.getInputStream());
        client.setSoTimeout(0);
    }
    
    /**
     * Fhrt die Startkommunikation mit dem Server durch und richtet den
     * ObjectOutputStream und ObjectInputStream ein.
     * @param id  Bytes, die zur Identifikation an den Server gesendet werden
     *            sollen.
     * @exception IOException  wenn whrend der Kommunikation ein Fehler
     *                auftritt.
     */
    protected void initCommunication(byte[] id) throws IOException {
        OutputStream out = socket.getOutputStream();
        InputStream  in  = socket.getInputStream();
        
        socket.setSoTimeout(CONNECT_TIMEOUT);
        out.write(id);
        byte answer = (byte) in.read();
        if (answer != CONNECT_OK) {
            throw new ConnectException((answer == BYE)
                                       ? "Verbindung wurde abgelehnt"
                                       : "Fehler beim Verbinden");
        }
        outStream = new ObjectOutputStream(out);
        outStream.flush();
        inStream  = new ObjectInputStream(in);
        socket.setSoTimeout(0);
    }

    /**
     * Wartet stndig auf Nachrichten und leitet sie an handleMessage(byte)
     * weiter. Bricht ab, wenn handleMessage(byte) false zurckliefert oder
     * wenn abortConnection aufgerufen wurde.
     */
    public void run() {
        try {
            while (handleMessage(inStream.readByte()));
        } catch (IOException ioe) {
            if (!aborted) {
                try {
                    handleMessage(CONNECTION_ERROR);
                } catch (IOException e) {}
            }
        }
    }

    /**
     * Auf eine Nachricht reagieren. NetworkSocket.handleMessage(byte) kennt nur
     * die Nachrichten BYE und CONNECTION_ERROR (Socket wird geschlossen und
     * false zurckgeliefert).
     * @param msgID  Die Identifikation der Nachricht, die empfangen wurde.
     * @return  true, wenn die Nachrichtenverarbeitung fortgesetzt werden soll.
     * @exception IOException  wenn die Nachricht unbekannt ist.
     */
    protected boolean handleMessage(byte msgID) throws IOException {
        switch (msgID) {
        case BYE:
        case CONNECTION_ERROR:
            socket.close();
            return false;
        default:
            throw new IOException("Unbekannte Nachricht");
        }
    }

    /**
     * Beendet die Verbindung. Es wird BYE gesendet, der Socket geschlossen und
     * Die Nachrichtenverarbeitung beendet.
     * @exception IOException  wenn beim Schlieen der Verbindung ein Fehler
     *                auftritt.
     */
    public synchronized void abortConnection() throws IOException {
        aborted = true;
        outStream.writeByte(BYE);
        outStream.flush();
        socket.close();
    }

    /**
     * Liefert den Socket fr diese Netzwerkverbindung.
     */
    public Socket getSocket() {
        return socket;
    }
    
    /**
     * Liefert den ObjectOutputStream fr diese Netzwerkverbindung.
     */
    protected ObjectOutputStream getOutStream() {
        return outStream;
    }

    /**
     * Liefert den ObjectInputStream fr diese Netzwerkverbindung.
     */
    protected ObjectInputStream getInStream() {
        return inStream;
    }
}
