/*
 * @(#)ClientServerSocket.java    14.6.1999
 */


package corewar.client;

import corewar.NetworkSocket;
import corewar.StatusField;
import corewar.GameSetupInfo;
import corewar.MessageBox;
import corewar.CoreInfo;
import java.net.InetAddress;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;

/**
 * Die Klasse ClientServerSocket stellt eine Netzwerkverbindung vom Client zum
 * Server dar. Sie ist fr die gesamte Kommunikation ds Clients mit dem Server
 * verantwortlich.
 *
 * @author  Ivo Kasiuk
 */
public class ClientServerSocket extends NetworkSocket {

    /* Die Netzwerkidentifikation fr sich direkt anmeldende Clients.  */
    private static final byte[] independentID = { I_AM_CLIENT, 0 };

    /* Der SimulationFrame, an den die Nachrichten weitergeleitet werden. */
    private SimulationFrame viewer;
    
    /**
     * Baut eine Verbindung zu einem Server auf und sendet ihm ein CoreInfo-
     * Objekt.
     * @param address  Die Adresse des Servers.
     * @param port     Der Port des Servers.
     * @param id       Die Identifikationsnummer fr diesen Client.
     * @param coreInfo Das CoreInfo-Objekt, das dem Server gesendet wird.
     * @exception IOException  Wenn bei der Anmeldung ein Fehler auftritt.
     */
    public ClientServerSocket(InetAddress address, int port, byte id,
                              CoreInfo coreInfo) throws IOException {
        super(address, port);
        byte[] clientID = { I_AM_CLIENT, id };
        initCommunication(clientID);
        sendCoreInfo(coreInfo);
    }

    /**
     * Baut als direkt anmeldender Client (ohne Verteiler) eine Verbindung zu
     * einem Server auf und sendet ihm ein CoreInfo-Objekt.
     * @param hostName  Der Hostname des Servers.
     * @param port      Der Port des Servers.
     * @param coreInfo  Das CoreInfo-Objekt, das dem Server gesendet wird.
     * @exception IOException  Wenn bei der Anmeldung ein Fehler auftritt.
     */
    public ClientServerSocket(String hostName, int port, CoreInfo coreInfo)
            throws IOException {
        super(hostName, port, independentID);
        sendCoreInfo(coreInfo);
    }

    /*
     * Versendet das CoreInfo-Objekt coreInfo.
     */
    private void sendCoreInfo(CoreInfo coreInfo) throws IOException {
        ObjectOutputStream out = getOutStream();
        out.writeByte(CORE_INFO);
        out.writeObject(coreInfo);
        out.flush();
    }

    /**
     * Wartet, bis der Server meldet, da die Spielrunde beginnt und empfngt
     * das GameSetupInfo-Objekt.
     * @param status  Statusfeld, in dem whrend des Wartens Statusinformationen
     *                angezeigt werden, falls es nicht den Wert null hat.
     * @return  das empfangene GameSetupInfo-Objekt.
     * @exception IOException  wenn bei der Kommunikation ein Fehler auftritt.
     */
    public GameSetupInfo waitForCores(StatusField status) throws IOException {
        ObjectInputStream in = getInStream();
        try {
            while (true) {
                byte msgID = in.readByte();
                switch (msgID) {
                case WAIT_FOR_CORES:
                    if (status != null) {
                        status.showText("Noch fehlende Cores: "
                                        + in.readByte());
                    }
                    break;
                case PARAM_INFO:
                    GameSetupInfo result = null;
                    try {
                        result = (GameSetupInfo) in.readObject();
                    } catch (Exception e) {}
                    if (result == null) {
                        throw new IOException("Fehler beim Empfangen der "
                                              + "Spieleinstellungen");
                    }
                    return result;
                case BYE:
                    throw new IOException("Verbindung vom Server abgebrochen");
                default:
                    throw new IOException("Fehler in der Netzwerkverbindung");
                }
            }
        } catch (IOException ioe) {
            getSocket().close();
            throw ioe;
        }
    }

    /**
     * Startet die Verarbeitung der ankommenden Nachrichten als eigenen Thread.
     * Die Nachrichten werden an die entsprechenden Methoden des
     * SimulationFrame-Objekts weitergeleitet. Vor dem Starten des Threads wird
     * dem Server signalisiert, da der Client bereit ist, die Spieldaten zu
     * empfangen.
     * @param viewer  Der SimulationFrame, an den die Nachrichten weitergeleitet
     *                werden sollen.
     * @exception IOException  wenn bei der Kommunikation ein Fehler auftritt.
     */
    public void start(SimulationFrame viewer) throws IOException {
        this.viewer = viewer;
        readyForData();
        super.start();
    }

    /**
     * Signalisiert dem Server, da der Client bereit ist fr die nchste
     * Spielrunde.
     * @exception IOException  wenn bei der Kommunikation ein Fehler auftritt.
     */
    public void readyForData() throws IOException {
        ObjectOutputStream out = getOutStream();
        out.writeByte(READY_FOR_DATA);
        out.flush();
    }
    
    /**
     * Reagiert auf die vom Server gesendeten Nachrichten. Die Nachrichten
     * werden an die entsprechenden Methoden des SimulationFrame-Objekts
     * weitergeleitet.
     * @param msgID  Die Identifikationsnummer der Nachricht.
     * @exception IOException  wenn bei der Kommunikation ein Fehler auftritt.
     */
    protected boolean handleMessage(byte msgID) throws IOException {
        ObjectInputStream in = getInStream();
        switch (msgID) {
        case RAM_INFO:
            int[] startPC = null;
            try {
                startPC = (int[]) in.readObject();
            } catch (Exception e) {}
            if (startPC == null) {
                throw new IOException("Fehler in der Netzwerkverbindung");
            }
            viewer.startRound(startPC);
            return true;
        case CYCLE_INFO:
            short[] pc           = null;
            short[] writeAddress = null;
            byte[]  writeLength  = null;
            try {
                pc           = (short[]) in.readObject();
                writeAddress = (short[]) in.readObject();
                writeLength  = (byte[]) in.readObject();
            } catch (Exception e) {}
            if ((pc == null) || (writeAddress == null)
                    || (writeLength == null)) {
                throw new IOException("Fehler in der Netzwerkverbindung");
            }
            viewer.handleCycle(pc, writeAddress, writeLength);
            return true;
        case END_OF_ROUND:
            if (viewer.endRound()) {
                getSocket().close();
                return false;
            }
            return true;
        case STOP:
        case BYE:
            getSocket().close();
            (new MessageBox("Spiel vom Server abgebrochen", "CoreWar", viewer,
                            MessageBox.OK)).execute();
            viewer.abortGame();
            return false;
        case CONNECTION_ERROR:
            (new MessageBox("Fehler in der Netzwerkverbindung", "CoreWar",
                            viewer, MessageBox.OK)).execute();
            viewer.abortGame();
            /* kein break */
        default:
            return super.handleMessage(msgID);
        }
    }
}

