/*
 * @(#)ConnectionDialog.java    27.6.1999
 */


package corewar.client;

import corewar.ToggleAction;
import corewar.StatusField;
import corewar.GameSetupInfo;
import corewar.ServerInfo;
import corewar.CoreInfo;
import java.awt.*;
import java.awt.event.WindowListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.io.IOException;
import java.util.Properties;

/**
 * Die Klasse ConnectionDialog stellt ein Dialogfenster zum Verbinden mit einem
 * Verteiler bzw. Server dar. Nach dem Schlieen des Dialogs kann das Ergebnis
 * mit getServerSocket() und getGameSetup() abgefragt werden.
 *
 * @author  Ivo Kasiuk
 */
public class ConnectionDialog extends Frame implements StatusField {

    /* Die Labels fr die "Verbinden"-Buttons. */
    private static final String[] connectLabels = { "Verbinden", "Stop" };

    /* Der Name des Properties fr den Verteilernamen. */
    private static final String PROPERTY_META_NAME = "metaServer";

    /* Der Name des Properties fr den Servernamen. */
    private static final String PROPERTY_SERVER_NAME = "server";

    /* Der Name des Properties fr den Spielernamen. */
    private static final String PROPERTY_PLAYER = "player";
    
    /* Das Textfeld fr die Verteileradresse. */
    private TextField metaText;

    /* Das Textfeld fr die Serveradresse. */
    private TextField serverText;

    /* Der "Verbinden"-Button fr den Verteiler. */
    private Button metaButton;

    /* Der "Verbinden"-Button fr den Server. */
    private Button serverButton;

    /* Das Textfeld fr den Spielernamen. */
    private TextField nameText;

    /* Das Statusfeld. */
    private TextArea  statusArea;

    /* Das ToggleAction-Objekt fr den "Verbinden"-Button fr den Verteiler. */
    private ToggleAction metaListener;

    /* Das ToggleAction-Objekt fr den "Verbinden"-Button fr den Server. */
    private ToggleAction serverListener;

    /* Der Core, mit dem der Client am Spiel teilnehmen will. */
    private CoreInfo coreInfo;

    /* Die Netzwerkverbindung zum Verteiler. */
    private ClientMetaSocket metaSocket = null;

    /* Die Netzwerkverbindung zum Server */
    private ClientServerSocket serverSocket = null;

    /* Das vom Server gesendete GameSetupInfo-Objekt */
    private GameSetupInfo gameSetup = null;

    /*
     * Signalisiert den Verbindungs-Threads, da das Verbinden abgebrochen
     * werden soll.
     */
    private boolean aborted = false;

    /**
     * Erzeugt und initialisiert das Dialogfenster.
     * @param coreInfo  Der Core, mit dem der Client am Spiel teilnehmen will.
     */
    public ConnectionDialog(CoreInfo coreInfo) {
        super("Verbinden");
        this.coreInfo = coreInfo;

        /* Unter Windows die Hintergrundfarbe korrigieren. */
        if (System.getProperty("os.name").startsWith("Windows")) {
            setBackground(SystemColor.control);
        }

        /*
         * Die zwei Panels connectPanel und corePanel bekommen ein GridBagLayout
         * und werden bereinander im Panel topPanel angeordnet.
         */
        Panel              topPanel     = new Panel(new BorderLayout());
        GridBagLayout      gridBag      = new GridBagLayout();
        Panel              connectPanel = new Panel(gridBag);
        Panel              corePanel    = new Panel(gridBag);
        GridBagConstraints constraints  = new GridBagConstraints();
        Properties         properties   = ClientProperties.getProperties();
        constraints.insets = new Insets(2, 2, 2, 2);
        constraints.fill   = GridBagConstraints.HORIZONTAL;

        /* Verteiler-Elemente in connectPanel einfgen. */
        connectPanel.add(makeLabel("Verteiler: ", gridBag, constraints));
        metaText = makeTextField(properties.getProperty(PROPERTY_META_NAME, ""),
                                 gridBag, constraints);
        connectPanel.add(metaText);
        metaButton   = makeButton("", gridBag, constraints);
        metaListener = new ToggleAction(metaButton, connectLabels, false) {
            public void buttonToggled(boolean activated) {
                if (activated) {
                    (new MetaConnectThread()).start();
                } else {
                    abortConnect();
                }
            }
        };
        connectPanel.add(metaButton);

        /* Server-Elemente in connectPanel einfgen. */
        connectPanel.add(makeLabel("Server: ", gridBag, constraints));
        serverText = makeTextField
                (properties.getProperty(PROPERTY_SERVER_NAME, ""),
                 gridBag, constraints);
        connectPanel.add(serverText);
        serverButton   = makeButton("", gridBag, constraints);
        serverListener = new ToggleAction(serverButton, connectLabels, false) {
            public void buttonToggled(boolean activated) {
                if (activated) {
                    (new ServerConnectThread()).start();
                } else {
                    abortConnect();
                }
            }
        };
        connectPanel.add(serverButton);

        /* corePanel fllen. */
        corePanel.add(makeLabel("Corename: ", gridBag, constraints));
        corePanel.add(makeTextLabel(coreInfo.coreName, gridBag, constraints));
        corePanel.add(makeLabel("Coregre: ", gridBag, constraints));
        corePanel.add(makeTextLabel(coreInfo.size + " bytes", gridBag,
                                    constraints));
        corePanel.add(makeLabel("Spieler: ", gridBag, constraints));
        if (coreInfo.author == null) {
            coreInfo.author = properties.getProperty(PROPERTY_PLAYER, "");
        }
        constraints.gridwidth = GridBagConstraints.REMAINDER;
        nameText = makeTextField(coreInfo.author, gridBag, constraints);
        corePanel.add(nameText);

        topPanel.add(connectPanel, "North");
        topPanel.add(corePanel, "South");
        add(topPanel, "North");

        /* Unter das topPanel kommt das Statusfeld. */
        statusArea = new TextArea(5, 30);
        statusArea.setEditable(false);
        add(statusArea, "Center");

        /* Der "Abbrechen"-Button kommt nach unten rechts. */
        Panel  buttonPanel  = new Panel(new FlowLayout(FlowLayout.RIGHT));
        Button cancelButton = new Button("Abbrechen");
        cancelButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                exitDialog();
            }
        });
        buttonPanel.add(cancelButton);
        add(buttonPanel, "South");
        
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {                
                exitDialog();
            }
        });
        pack();
        centerScreen();
    }

    private void centerScreen() {
        Dimension dim = getToolkit().getScreenSize();
        Rectangle abounds = getBounds();
        setLocation((dim.width - abounds.width) / 2,
                    (dim.height - abounds.height) / 2);
    }

    /*
     * Erstellt ein Label, das eine Beschriftung darstellt.
     */
    private Label makeLabel(String name, GridBagLayout gridBag,
                            GridBagConstraints constraints) {
        Label result = new Label(name, Label.RIGHT);
        constraints.gridwidth = 1;
        constraints.anchor    = GridBagConstraints.EAST;
        gridBag.setConstraints(result, constraints);
        return result;
    }

    /*
     * Erstellt ein Label, das eine Information auf dem Bildschirm anzeigt.
     */
    private Label makeTextLabel(String text, GridBagLayout gridBag,
                                GridBagConstraints constraints) {
        Label result = new Label(text, Label.LEFT);
        constraints.gridwidth = GridBagConstraints.REMAINDER;
        constraints.anchor    = GridBagConstraints.WEST;
        gridBag.setConstraints(result, constraints);
        return result;
    }

    /*
     * Erstellt ein Textfeld mit variabler Breite.
     */
    private TextField makeTextField(String text, GridBagLayout gridBag,
                                    GridBagConstraints constraints) {
        TextField result = new TextField(text);
        constraints.weightx = 1.0;
        gridBag.setConstraints(result, constraints);
        return result;
    }

    /*
     * Erstellt eine Button mit fester Breite.
     */
    private Button makeButton(String label, GridBagLayout gridBag,
                              GridBagConstraints constraints) {
        Button result = new Button(label);
        constraints.weightx   = 0.0;
        constraints.gridwidth = GridBagConstraints.REMAINDER;
        gridBag.setConstraints(result, constraints);
        return result;
    }

    /*
     * Beendet den Dialog. Vorher wird ggf. der Verbindungsvorgang abgebrochen
     * und die Einstellungen in ClientProperties gespeichert.
     */
    private void exitDialog() {
        abortConnect();
        Properties properties = ClientProperties.getProperties();
        properties.put(PROPERTY_META_NAME, metaText.getText());
        properties.put(PROPERTY_SERVER_NAME, serverText.getText());
        properties.put(PROPERTY_PLAYER, nameText.getText());
        dispose();
    }
    
    /*
     * Bricht den Verbindungsvorgang ab.
     */
    private synchronized void abortConnect() {
        aborted = true;
        if ((metaSocket == null) && (serverSocket == null)) {
            return;
        }
        metaButton.setEnabled(false);
        serverButton.setEnabled(false);
        
        if (metaSocket != null) {
            try {
                metaSocket.abortConnection();
            } catch (IOException ioe) {}
        }
        if (serverSocket != null) {
            try {
                serverSocket.abortConnection();
            } catch (IOException ioe) {}
        }
    }

    /*
     * Thread zur Anmeldung bei einem Verteiler
     */
    private class MetaConnectThread extends Thread {

        /*
         * Graut den "Verbinden"-Button fr den Server ab.
         */
        public MetaConnectThread() {
            serverButton.setEnabled(false);
            aborted = false;
        }

        /*
         * Fhrt die Anmeldung bei einem Verteiler durch und bernimmt auch die
         * darauf folgende Anmeldung beim Server.
         */
        public void run() {
            coreInfo.author = nameText.getText();
            
            /* Beim Verteiler anmelden. */
            try {
                synchronized (ConnectionDialog.this) {
                    if (aborted) {    /* "Stop" oder "Abbrechen" gedrckt? */
                        throw new IOException();
                    }
                    showText("Verbindung zum Verteiler wird aufgebaut...");
                    metaSocket = new ClientMetaSocket(metaText.getText(),
                                                      coreInfo);
                }
            } catch (UnknownHostException uhe) {
                handleConnectError("Unbekannter Server: " + uhe.getMessage());
                return;
            } catch (IOException ioe) {
                handleConnectError("Verbinden fehlgeschlagen: "
                                   + ioe.getMessage());
                return;
            }

            /* Warten, bis ein Server zugewiesen wird. */
            ServerInfo serverInfo;
            try {
                showText("Bitte warten, bis ein Server bereit ist...");
                serverInfo = metaSocket.waitForServer();
            } catch (IOException ioe) {
                handleConnectError("Verbindung abgebrochen: "
                                   + ioe.getMessage());
                return;
            }

            /* Beim Server anmelden. */
            try {
                synchronized (ConnectionDialog.this) {
                    if (aborted) {    /* "Stop" oder "Abbrechen" gedrckt? */
                        throw new IOException();
                    }
                    showText("Verbindung zum Server wird aufgebaut...");
                    serverSocket = new ClientServerSocket (serverInfo.address,
                             serverInfo.port, serverInfo.id, coreInfo);
                }
                waitForCores();
            } catch (IOException ioe) {
                handleConnectError("Verbinden fehlgeschlagen: "
                                   + ioe.getMessage());
            }
        }
    }

    /*
     * Thread zur direkten Anmeldung bei einem Server.
     */
    private class ServerConnectThread extends Thread {

        /*
         * Graut den "Verbinden"-Button fr den Verteiler ab.
         */
        public ServerConnectThread() {
            metaButton.setEnabled(false);
            aborted = false;
        }

        /*
         * Fhrt die Direktanmeldung bei einem Server durch. Es werden ggf. alle
         * fr den Server vorgesehen Ports durchprobiert.
         */
        public void run() {
            coreInfo.author = nameText.getText();
            String hostName = serverText.getText();
            for (int i=ClientServerSocket.DEFAULT_SERVER_PORT;
                    i<(ClientServerSocket.DEFAULT_SERVER_PORT
                       + ClientServerSocket.MAX_SERVER_PORTS); i++) {
                try {
                    synchronized (ConnectionDialog.this) {
                        if (!aborted) {   /* "Stop" oder "Abbrechen" gedrckt? */
                            statusArea.append("Suche Server an Port " + i + "...");
                            serverSocket = new ClientServerSocket(hostName, i,
                                                                  coreInfo);
                            showText("gefunden.");
                        }
                    }
                    break;
                } catch (UnknownHostException uhe) {
                    statusArea.append("\n");
                    handleConnectError("Unbekannter Server: "
                                       + uhe.getMessage());
                    return;
                } catch (IOException ioe) {
                    showText("fehlgeschlagen: " + ioe.getMessage());
                }
            }

            if (serverSocket == null) {
                handleConnectError("Kein Server gefunden.");
            } else {
                waitForCores();
            }
        }
    }
    
    /*
     * Wartet, bis das Spiel beginnt und beendet dann den Dialog.
     */
    private void waitForCores() {
        showText("Warte auf andere Cores...");
        try {
            gameSetup = serverSocket.waitForCores(this);
            dispose();
        } catch (IOException ioe) {
            handleConnectError("Verbindung abgebrochen: " + ioe.getMessage());
        }
    }
    
    /*
     * Macht nach einer fehlgeschlagenen Verbindung den Dialog wieder fr einen
     * erneuten Verbindungsversuch bereit. Auerdem wird die Fehlermeldung
     * errorMsg im Statusfeld angezeigt.
     */
    private synchronized void handleConnectError(String errorMsg) {
        showText(aborted ? "Verbindungsvorgang abgebrochen."
                         : errorMsg);
        metaSocket   = null;
        serverSocket = null;
        metaListener.setActivated(false);
        metaButton.setEnabled(true);
        serverListener.setActivated(false);
        serverButton.setEnabled(true);
    }
    
    /**
     * Zeigt das Dialogfenster an.
     * @param listener  Wird als WindowListener fr den Dialog eingetragen. 
     */
    public void show(WindowListener listener) {
        addWindowListener(listener);
        setVisible(true);
    }

    /**
     * Liefert die Netzwerkverbindung zum Server.
     */
    public ClientServerSocket getServerSocket() {
        return (gameSetup == null) ? null : serverSocket;
    }

    /**
     * Liefert das vom Server gesendete GameSetupInfo-Objekt.
     */
    public GameSetupInfo getGameSetup() {
        return gameSetup;
    }
    
    /**
     * Zeigt einen Text im Statusfeld des Dialogfensters an.
     */
    public void showText(String text) {
        statusArea.append(text + '\n');
    }

    /**
     * Lscht den Inhalt des Statusfeldes.
     */
    public void clear() {
        statusArea.setText("");
    }
    
    public static void main(String[] args) throws IOException {
        try {
            ClientProperties.load();
	} catch (IOException ioe) {}
        CoreInfo core = new CoreInfo();
        core.coreName = "Super Core";
        core.size     = 346;
        ConnectionDialog dialog = new ConnectionDialog(core);
        dialog.show(new WindowAdapter() {
            public void windowClosed(WindowEvent event) {
                try {
                    ClientProperties.save();
                } catch (IOException ioe) {}
                System.exit(0);
            }
        });
    }
}
