/*
 * @(#)HistoryFrame.java    17.6.1999
 */


package corewar.client;

import corewar.GameSetupInfo;
import corewar.ParamInfo;
import corewar.GameInfo;
import corewar.CoreInfo;
import java.awt.*;
import java.awt.event.WindowListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.AdjustmentEvent;
import java.io.File;
import java.io.IOException;
import java.util.Vector;

/**
 * Die Klasse HistoryFrame ermglicht das schrittweise Betrachten einer
 * gespeicherten Spielrunde.
 *
 * @author  Ivo Kasiuk
 */
public class HistoryFrame extends ViewerFrame {

    /* Der Name des Properties fr die Verzgerung zwischen zwei Zyklen. */
    private static final String PROPERTY_DELAY = "historyDelay";
    
    /* Der Scroller, der die aktuelle Position in der History darstellt. */
    private Scrollbar cycleScroll;

    /* Die Abspielrichtung (1=vorwrts, 0=stop, -1=rckwrts). */
    private byte direction = 0;

    /* Verzgerung zwischen zwi Zyklen int ms. */
    private int delay = 0;

    /* Zyklus, bis zu dem die History vorwrts abgespielt werden soll. */
    private int maxCycle;

    /* Zyklus, bis zu dem die History rkckwrts abgespielt werden soll. */
    private int minCycle = 0;

    /* Der Thread zum Abspielen der History. */
    private PlayThread playThread;

    /* Gibt an, ob der HistoryFrame beendet wurde. */
    private boolean aborted = false;
    
    /**
     * Initialisiert den HistoryFrame mit den Daten einer gespeicherten
     * Spielrunde.
     * @param historyFile  Die History-Datei.
     * @exception IOException  wenn beim Laden ein Fehler auftritt.
     */
    public HistoryFrame(File historyFile/*GameSetupInfo gameSetup*/)
            throws IOException, InterruptedException {
        super(historyFile/*gameSetup*/);

        GridBagLayout      gridBag     = new GridBagLayout();
        GridBagConstraints constraints = new GridBagConstraints();
        toolPanel.setLayout(gridBag);
        constraints.insets    = new Insets(0, 3, 0, 3);
        constraints.fill      = GridBagConstraints.HORIZONTAL;
        constraints.weightx   = 1.0;
        constraints.gridwidth = GridBagConstraints.REMAINDER;

        Label cycleLabel = new Label("Zyklus:", Label.CENTER);
        gridBag.setConstraints(cycleLabel, constraints);
        toolPanel.add(cycleLabel);
        
        maxCycle = status.getGameInfo().getMaxCycles();
        cycleScroll = new Scrollbar(Scrollbar.HORIZONTAL, 0, maxCycle / 10, 0,
                                       maxCycle / 10 + maxCycle);
        cycleScroll.setBlockIncrement(maxCycle / 100);
        cycleScroll.addAdjustmentListener(new AdjustmentListener() {
            public void adjustmentValueChanged(AdjustmentEvent event) {
                goToCycle(event.getValue());
            }
        });
        gridBag.setConstraints(cycleScroll, constraints);
        toolPanel.add(cycleScroll);

        Label speedLabel = new Label("Geschwindigkeit:", Label.CENTER);
        gridBag.setConstraints(speedLabel, constraints);
        toolPanel.add(speedLabel);
        
        String delayString = ClientProperties.getProperties()
                .getProperty(PROPERTY_DELAY);
        if (delayString != null) {
            try {
                delay = Integer.parseInt(delayString);
            } catch (NumberFormatException nfe) {}
        }
        Scrollbar speedScroll = new Scrollbar(Scrollbar.HORIZONTAL,
                                              (1000 - delay) / 10, 10, 0, 110);
        speedScroll.addAdjustmentListener(new AdjustmentListener() {
            public void adjustmentValueChanged(AdjustmentEvent event) {
                delay = (100 - event.getValue()) * 10;
            }
        });
        gridBag.setConstraints(speedScroll, constraints);
        toolPanel.add(speedScroll);
        
        Panel buttonPanel = new Panel(new GridLayout(1, 0, 3, 0));
        Button resetButton = new Button("|<<");
        resetButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                resetHistory();
            }
        });
        buttonPanel.add(resetButton);
        
        Button leftButton = new Button("<");
        leftButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                playBackward();
            }
        });
        buttonPanel.add(leftButton);
        
        Button stopButton = new Button("Stop");
        stopButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                stopPlaying();
            }
        });
        buttonPanel.add(stopButton);
        
        Button rightButton = new Button(">");
        rightButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                playForward();
            }
        });
        buttonPanel.add(rightButton);
        
        constraints.insets = new Insets(5, 3, 10, 3);
        gridBag.setConstraints(buttonPanel, constraints);
        toolPanel.add(buttonPanel);
        
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                aborted = true;
                ClientProperties.getProperties().put(PROPERTY_DELAY,
                                                     String.valueOf(delay));
                dispose();
            }
        });
        pack();

        //playThread = new PlayThread();
        //playThread.start();
    }

    public void loadHistory() throws IOException, InterruptedException {
        super.loadHistory();
        resetHistory();
        playThread = new PlayThread();
        playThread.start();
    }

    /*
     * Spielt die History vorwrts ab.
     */
    private synchronized void playForward() {
        direction = 1;
        maxCycle  = status.getGameInfo().getMaxCycles();
        playThread.resume();
    }

    /*
     * Spielt die History rckwrts ab.
     */
    private synchronized void playBackward() {
        direction = -1;
        minCycle  = 0;
        playThread.resume();
    }

    /*
     * Spielt die History bis zu einem bestimmten Zyklus ab.
     */
    private synchronized void goToCycle(int destCycle) {
        if (destCycle < status.getGameInfo().getCycleNumber()) {
            direction = -1;
            minCycle  = destCycle;
        } else {
            direction = 1;
            maxCycle  = destCycle;
        }
        playThread.resume();
    }
    
    /*
     * Hlt das Abspielen an.
     */
    private synchronized void stopPlaying() {
        direction = 0;
    }

    /*
     * History an den Anfang zurcksetzen.
     */
    private synchronized void resetHistory() {
        stopPlaying();
        history.reset();
        cycleScroll.setValue(0);
        status.setCycle(0);
        status.setActiveCores(status.getGameInfo().getCoreNumber());
        Vector cores = coreList.getElements();
        for (int i=0; i<cores.size(); i++) {
            ((CoreStatus) cores.elementAt(i)).alive = true;
        }
        coreList.refreshAll();
    }

    /*
     * Spielt einen Zyklus vorwrts ab.
     */
    private synchronized void playCycleForward() {
        cycleForward();
        GameInfo gameInfo = status.getGameInfo();
        if (maxCycle == gameInfo.getMaxCycles()) {
            cycleScroll.setValue(gameInfo.getCycleNumber());
        }
    }

    /*
     * Spielt einen Zyklus rckwrts ab.
     */
    private synchronized void playCycleBackward() {
        cycleBackward();
        if (minCycle == 0) {
            cycleScroll.setValue(status.getGameInfo().getCycleNumber());
        }
    }
    
    /*
     * Der Thread zum Abspielen der History.
     */
    private class PlayThread extends Thread {

        /*
         * Spielt die History ab und richtet sich dabei nach den Variablen
         * direction, delay, maxCycle und minCycle.
         */
        public void run() {
            while (!aborted) {
                switch (direction) {
                case -1:
                    if (status.getGameInfo().getCycleNumber() <= minCycle) {
                        stopPlaying();
                    } else {
                        playCycleBackward();
                        try {
                            sleep(delay);
                        } catch (InterruptedException ie) {}
                    }
                    break;
                case 0:
                    suspend();
                    break;
                case 1:
                    if (status.getGameInfo().getCycleNumber() >= maxCycle) {
                        stopPlaying();
                    } else {
                        playCycleForward();
                        try {
                            sleep(delay);
                        } catch (InterruptedException ie) {}
                    }
                    break;
                default:
                    System.err.println("Ungltiger Wert fr direction: "
                                       + direction + '.');
                    return;
                }
            }
        }
    }

/*    
    public static void main(String[] args) {
        try {
            ClientProperties.load();
        } catch (IOException ioe) {}
        ParamInfo param = new ParamInfo();
        param.setRAMSize(8192);
        param.setMaxCycles(10000);
        Vector cores = new Vector();
        CoreInfo core = new CoreInfo();
        core.coreName = "Super Core";
        core.author = "Coremaster";
        core.size = 364;
        cores.addElement(core);
        core = new CoreInfo();
        core.coreName = "Raupe";
        core.author = "Anonymous";
        core.size = 6;
        cores.addElement(core);
        core = new CoreInfo();
        core.coreName = "Dabeiseinisalles";
        core.author = "Loser";
        core.size = 142;
        cores.addElement(core);
        for (int i=0; i<17; i++) {
            core = new CoreInfo();
            core.coreName = "TestCore";
            core.author = "TestAutor";
            core.size = 123;
            cores.addElement(core);
        }
        GameSetupInfo setup = new GameSetupInfo(param, cores, (byte) 0);
        try {
            HistoryFrame viewer = new HistoryFrame(setup);
            viewer.show(new WindowAdapter() {
                public void windowClosed(WindowEvent event) {
                    try {
                        ClientProperties.save();
                    } catch (IOException ioe) {}
                    System.exit(0);
                }
            });
        } catch (IOException ioe) {}
    }
    */
}
