/*
 * @(#)RAMDisplay.java   10.07.1999
 */
 
package corewar.client;

import java.awt.*;
import java.awt.image.MemoryImageSource;
import java.util.Random;
import java.util.Vector;

/**
 * Die Klasse ermglicht eine visuelle Anzeige des Spielspeichers, dabei werden
 * die Cores mit Farben gekennzeichnet.
 *
 * @author Hendrick Kay
 */
public class RAMDisplay extends Canvas {

    /** @serial SpeicherArray mit CoreNummern */
    private byte[] ram;

    /** @serial Vector mit CoreStatus-Objekten */
    private Vector cores;

    /** @serial Speicherpixel-Breite */
    private int pixelWidth;
    
    /** @serial Speicher-Hhe */
    private int pixelHeight;

    /** @serial Anzahl der Spalten in der Anzeige */
    private int columns;

    /** @serial Zur Beschleunigung der Ausgabe */
    private short xMaskAND;

    /** @serial Zur Beschleunigung der Ausgabe */
    private short xDivisionShift;

    /** @serial Zur Beschleunigung der Ausgabe */
    private short yDivisionShift;

    /** @serial Gre des Speichers */
    private int coreSize;

    /** @serial Breite und Hhe dieses Objekts */
    private final static int XYSIZE = 512;

    // Die nchsten 3 Felder werden fr das DoubleBuffering bentigt

    /** @serial Bildschirm-Buffer */
    private Image doubleBuffer = null;

    /** @serial Dimension des Canvas */
    private Dimension canvasDimension;

    /** @serial Graphikkontext des Images */
    private Graphics bufferGraphics;

    /**
     * Initialsiert den Speicher
     *
     * @param coreSize  Gre des Spielspeichers
     * @param cores     Vector mit allen spielenden Cores
     */
    public RAMDisplay(int coreSize, Vector cores) {

        // Dimension des Canvas einrichten
        canvasDimension = new Dimension(XYSIZE, XYSIZE);

        // Instanzvariablen setzen
        this.cores = cores;
        this.coreSize = coreSize;

        // neues ramArray erzeugen
        ram = new byte[coreSize];

        // alle Stellen im RAM als besitzerlos (-1) markieren
        for (int i=0; i<coreSize; i++) {
            ram[i] = (byte) -1;
        }
    
        // Gre der Anzeige setzen
        setSize(XYSIZE, XYSIZE);

        // Hintergrund auf schwarz setzen
        setBackground(Color.black);


        // SpeicherPixel-Gren und shiftWerte berechnen 
        pixelWidth     = 32;
        pixelHeight    = 32;
        xDivisionShift = 5;        // 2^5=32
        yDivisionShift = 5;        // 2^5=32
        do {

            // prfen, ob mit den aktuellen Pixelbreiten alles ins Display pat
            if ( ( XYSIZE/pixelWidth ) * ( XYSIZE/pixelHeight) < coreSize ) {
                // Speicherpixelbreite halbieren
                pixelWidth >>= 1;
                xDivisionShift--;
            }
            else {
                // Der Speicher pat in die Anzeige
                break;
            }

            // prfen, ob mit den aktuellen Pixelbreiten alles ins Display pat
            if ( ( XYSIZE/pixelWidth ) * ( XYSIZE/pixelHeight) < coreSize ) {
                // Speicherpixelhhe halbieren
                pixelHeight >>= 1;
                yDivisionShift--;

            }
            else {
                // Der Speicher pat in die Anzeige
                break;
            }

        } while ( true );

        // Anzahl der Spalten berechnen
        columns = XYSIZE/pixelWidth;

        // UND-Maske berechnen
        xMaskAND = 0;
        while ( xMaskAND < columns-1) {
            xMaskAND <<= 1;
            xMaskAND++;
        }
        
    } // RAMDisplay()

    /**
     * Setzt die Start-PCs fr alle Cores
     *
     */
    public void setStartPC(int[] startPC) {
        CoreStatus cs;
        for (byte i = 0; i < cores.size() ;i++) {
            cs = (CoreStatus) cores.elementAt(i);
            cs.pc = startPC[i];
        }
    }

    /**
     * Kopiert die Cores in den Speicher
     *
     */
    public void copyCorestoRAMDisplay() {
        System.out.println("copyCoresTo... called");
       // CoreCode in den Speicher schreiben
        CoreStatus cs;
        int address;
        for (byte i = 0; i < cores.size() ;i++) {
            cs = (CoreStatus) cores.elementAt(i);
            address = cs.pc;
            for ( int n = 0; n < cs.size ;n++) {
                try {
                    ram[address+n] = i;
                }
                catch (ArrayIndexOutOfBoundsException e) {
                    ram[(address+n)%coreSize] = i;
                }
            }
        }
     }

    /**
     * Liefert die minimale Gre des Canvas
     *
     * @return Die minimale Gre.
     */
    public synchronized Dimension getMinimumSize() {
        return canvasDimension;
    } // getMinimumSize()

    /**
     * Liefert die gewnschte Gre fr den Canvas.
     *
     * @return Die gewnschte Gre.
     */
    public Dimension getPreferredSize() {
        return getMinimumSize();
    } // getPreferredSize()

    /**
     * Setzt ein Speicherpixel auf eine Corefarbe.
     *
     * @param coreNumber  Nummer des Cores. 0 setzt wieder die Hintergrundfarbe.
     * @param address     Addresse.
     */
    public synchronized void setByte(byte coreNumber, int address) {

        // DoubleBuffering schon aktiv?
        if ( doubleBuffer == null ) initDoubleBuffering();
       
        // Grafikkontext vom Canvass holen
        Graphics g = getGraphics();


        // Farbe des Cores ermitteln und in Canvas und Buffer setzen
        Color color;
        if (coreNumber == -1) {
            color = Color.black;
        }
        else {
            color = ((CoreStatus) cores.elementAt(coreNumber) ).color;
        }

        g.setColor( color );
        bufferGraphics.setColor( color );

        // Speicherpixel schreiben
        setPoint( g, address );
        setPoint( bufferGraphics, address );

        try {
            ram[address] = coreNumber;
        }
        catch ( ArrayIndexOutOfBoundsException e) {
            ram[address%coreSize] = coreNumber;
        }
    } // setByte()

    /**
     * Setzt mehrere Speicherpixel auf eine Corefarbe.
     *
     * @param coreNumber  Nummer des Cores.
     * @param address     Addresse.
     * @param length      Schreiblnge.
     */
    public synchronized void setBytes(byte coreNumber, int address, short length) {

        // DoubleBuffering schon aktiv?
        if ( doubleBuffer == null ) initDoubleBuffering();

        // Grafikkontext vom Canvas holen
        Graphics g = getGraphics();

        Color color;
        if (coreNumber == -1) {
            color = Color.black;
        }
        else {
            // Farbe des Cores ermitteln und in Canvas und Buffer setzen
            color = ((CoreStatus) cores.elementAt(coreNumber) ).color;
        }

        g.setColor( color );
        bufferGraphics.setColor( color );

        // Speicherpixel schreiben
        for (int i=0; i<length;i++) {
            setPoint( g, address+i );
            setPoint( bufferGraphics, address+i );
            try {
                ram[(address+i)%coreSize] = coreNumber;
            }
            catch ( ArrayIndexOutOfBoundsException e) {
                ram[(address+i)%coreSize] = coreNumber;
            }
        }
    } // setBytes()

    /**
     * Liefert die Corenummer an Speicherstelle address.
     *
     * @return Die Corenummer an der Speicherstelle.
     */
    public byte getByte(int address) {
        try {
            return ram[address];
        }
        catch (ArrayIndexOutOfBoundsException e) {
            return ram[address%coreSize];
        }
    } // getByte()

    /**
     * Setzt den PC eines Cores neu.
     *
     * @param coreNumber  Nummer des Cores
     * @param pc          neuer PC
     */
    public synchronized void setPC(byte coreNumber, int pc) {
        Graphics g = getGraphics();

        // alten PC holen
        int oldPC = ( (CoreStatus) cores.elementAt(coreNumber) ).pc;

        // nur entfernen, wenn zwischenzeitlich kein anderer geschrieben hat
        if ( coreNumber == getByte(oldPC) ) {

            // Farbe des Cores holen
            g.setColor( ( (CoreStatus) cores.elementAt(coreNumber) ).color );
            // alten PC entfernen
            setPoint( g, oldPC );

        }
        else {

            byte toll = getByte(oldPC);

            if ( toll == -1 ) {
                g.setColor( Color.black );
            }
            else {
                // Farbe des Cores holen
                g.setColor( ( (CoreStatus) cores.elementAt(toll) ).color );
            }
            // alten PC entfernen
            setPoint( g, oldPC );
        }

        // neuen PC im Bild darstellen
        g.setColor( Color.white );
        setPoint( g, pc );

        // neuen PC speichern
        ( (CoreStatus) cores.elementAt(coreNumber) ).pc = pc;


    } // setPC

    /**
     * Setzt die Anzeige wieder in den Startzustand zurck.
     *
     */
    public void reset() {

        // alle Stellen im RAM als besitzerlos (-1) markieren
        for (int i=0; i<coreSize; i++) {
            ram[i] = (byte) -1;
        }

        // CoreCode in den Speicher schreiben
        copyCorestoRAMDisplay();

        // DoubleBuffer updaten
        copyRAMtoBuffer();

        // alles neu zeichnen
        repaint();

    }

    /**
     * Bringt den Speicherinhalt zur Anzeige.
     *
     * @param g Grafikkontext
     */
    public synchronized void paint(Graphics g) {

        if ( doubleBuffer == null ) initDoubleBuffering();

        // Hintergrund lschen
        g.setColor( Color.black );
        g.fillRect(0, 0, XYSIZE, XYSIZE);

        // Speicher aus dem Buffer zeichnen
        g.drawImage(doubleBuffer, 0, 0, this);

        // PCs aller Cores zeichnen
        g.setColor( Color.white );
        for (int core=0; core < cores.size(); core++ ) {
            setPoint( g, ((CoreStatus) cores.elementAt(core) ).pc );
        }
    } // paint()

    /**
     * Setzt einen Speicherpixel in der aktuellen Farbe.
     *
     * @param g        Grafikkontext
     * @param address  Adresse die beschrieben werden soll.
     */
    private void setPoint(Graphics g, int address) {
        if (address >= ram.length) {
            address %= coreSize;
        }
        g.fillRect( (address & xMaskAND) << xDivisionShift,
                    (address / columns) << yDivisionShift, pixelWidth, pixelHeight);
        
    } // setPoint()

    /**
     * Initialisiert das DoubleBuffering zur Beschleunigung der Grafikausgabe.
     *
     */
    private void initDoubleBuffering() {
        // einen Buffer erzeugen
        doubleBuffer = createImage(XYSIZE,XYSIZE);

        // Grafikkontext fr Buffer erzeugen
        bufferGraphics = doubleBuffer.getGraphics();

        // Hintergrund des Buffers auf schwarz setzen
        bufferGraphics.setColor(Color.black);
        bufferGraphics.fillRect(0, 0, XYSIZE, XYSIZE);

        // Inhalt des Speichers in den Buffer kopiern (etwas aufwendig!)
        copyRAMtoBuffer();

    } // initDoubleBuffering()

    /**
     * kopiert den Inhalt des Spielspeichers in den Buffer
     *
     */
    private void copyRAMtoBuffer() {

        // DoubleBuffering schon aktiv?
        if ( doubleBuffer == null ) initDoubleBuffering();

        // ganzen Speicher in Buffer zeichnen
        for (int address=0; address<coreSize; address++) {

            // Farbe des Cores holen
            Color color;
            if ( ram[address] == -1) {
                color = Color.black;
            }
            else {
                color = ( (CoreStatus) cores.elementAt(ram[address]) ).color;
            }

            // Stiftfarbe setzen
            bufferGraphics.setColor( color );

            // Speicherpixel in Buffer zeichnen
            bufferGraphics.fillRect( (address & xMaskAND) << xDivisionShift,
                (address / columns) << yDivisionShift, pixelWidth, pixelHeight);
        }
    } // copyRAMtoBuffer()

} // RAMDisplay
