/*
 * @(#)ViewerFrame.java    8.7.1999
 */


package corewar.client;

import corewar.*;
import java.awt.*;
import java.awt.event.WindowListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.util.Vector;

/**
 * Die Klasse ViewerFrame stellt ein Grundgerst fr die Anzeige eines CoreWar-
 * Spiels dar.
 *
 * @author  Ivo Kasiuk
 */
public class ViewerFrame extends Frame {

    /*
     * Die Farben fr die Cores. Die erste Farbe ist fr den eigenen Core
     * reserviert.
     */
    private static final Color[] coreColors = { Color.red, Color.blue,
            new Color(0, 230, 255), Color.gray, new Color(0, 230, 0),
            Color.lightGray, Color.magenta, Color.orange, Color.yellow,
            new Color(170, 140, 0), new Color(150, 0, 150),
            new Color(0, 150, 150), new Color(255, 150, 150),
            new Color(150, 255, 150), new Color(150, 150, 255),
            new Color(160, 0, 0), new Color(50, 130, 0),
            new Color(200, 170, 110), new Color(220, 90, 150),
            new Color(240, 150, 30) };

    /** Das StepHistory-Objekt, das alle Spieldaten verwaltet. */
    protected StepHistory history;

    /** Die Liste aller am Spiel beteiligten Cores. */
    protected ClientCoreList coreList;

    /** Das Anzeigefeld fr das Spielgeschehen. */
    protected RAMDisplay ram;

    /** Das Anzeigefeld fr Spielinformationen. */
    protected StatusDisplay status;

    /**
     * In diesem Panel knnen abgeleitete Klassen zustzliche Bedienelemente
     * anordnen.
     */
    protected Panel toolPanel;

    /** Gibt an, ob gerade eine Spielrunde luft. */
    protected boolean playing = false;

    /** Die Anzahl der laufenden Speicherthreads. */
    protected int saveThreads = 0;

    /* History-Datei */
    private ObjectInputStream inStream;
    
    /* Name der History-Datei */
    private String fileName;

    /* Statusanzeige whrend des Ladens */
    private LoadBox loadBox;

    /**
     * Erzeugt und initialisiert den ViewerFrame.
     * @param gameSetup  Die Spieleinstellungen fr das CoreWar-Spiel.
     */
    public ViewerFrame(GameSetupInfo gameSetup) {
        super("CoreWar");
        
        Vector cores     = new Vector();
        Vector coreInfos = gameSetup.cores;
        byte   numCores  = (byte) coreInfos.size();
        int    count = 1;
        for (int i=0; i<numCores; i++) {
            CoreStatus core = new CoreStatus((CoreInfo) coreInfos.elementAt(i));
            if (i == gameSetup.selfIndex) {
                core.color = coreColors[0];
            } else {
                core.color = coreColors[count++];
            }
            core.score = 0;
            cores.addElement(core);
        }

        GameInfo  gameInfo  = new GameInfo();
        ParamInfo paramInfo = gameSetup.paramInfo;
        gameInfo.setCoreNumber(numCores);
        gameInfo.setRAMSize(paramInfo.getRAMSize());
        gameInfo.setMaxCycles(paramInfo.getMaxCycles());
        gameInfo.setTotalRounds(paramInfo.getTotalRounds());
        gameInfo.setCycleNumber(0);
        gameInfo.setRoundNumber((byte) 0);

        initFrame(gameInfo, cores);
    }

    /**
     * Ldt die Spieleinstellungen einer gespeicherten Spielrunde.
     * Die Spieldaten mssen danach mit loadHistory() geladen werden.
     * @param historyFile  Die History-Datei.
     * @exception IOException           wenn beim Laden ein Fehler auftritt.
     * @exception InterruptedException  wenn das Laden vom Benutzer abgebrochen
     *                                  wird.
     */
    public ViewerFrame(File historyFile)
            throws IOException, InterruptedException {
        super("History");
        inStream = new ObjectInputStream(new FileInputStream(historyFile));
        fileName = historyFile.getName();
        loadBox  = new LoadBox(fileName, inStream);
        loadBox.setVisible(true);

        try {
            GameInfo gameInfo = null;
            try {
                gameInfo = (GameInfo) inStream.readObject();
            } catch (Exception e) {}
            if (gameInfo == null) {
                throw new IOException(fileName + " konnte nicht geladen "
                                      + "werden.");
            }

            Vector cores = new Vector();
            for (int i=0; i<gameInfo.getCoreNumber(); i++) {
                CoreStatus core = null;
                try {
                    core = (CoreStatus) inStream.readObject();
                } catch (Exception e) {}
                if (core == null) {
                    throw new IOException(fileName + " konnte nicht geladen "
                                          + "werden.");
                }
                cores.addElement(core);
            }
            
            initFrame(gameInfo, cores);
            setCoresAlive();
        } catch (IOException ioe) {
            LoadBox box = loadBox;
            abortLoading();
            if (box.getAborted()) {
                throw new InterruptedException("Ladevorgang abgebrochen");
            }
            throw ioe;
        }
    }
    
    /**
     * Ldt die Spieldaten einer gespeicherten Spielrunde und startet sie.
     * Das Objekt muss mit ViewerFrame(File) angelegt worden sein.
     * @exception IOException           wenn beim Laden ein Fehler auftritt.
     * @exception InterruptedException  wenn das Laden vom Benutzer abgebrochen
     *                                  wird.
     */
    public void loadHistory() throws IOException, InterruptedException {
        //try {
            history = new StepHistory(inStream, ram,
                                      status.getGameInfo().getMaxCycles());
            playing = true;
        //} catch (IOException ioe) {
        //    if (loadBox.getAborted()) {
        //        throw new InterruptedException("Ladevorgang abgebrochen");
        //    }
        //    throw ioe;
        //} finally {
            abortLoading();
        //}    
    }

    /*
     * Bricht den Ladevorgang ab.
     */
    private void abortLoading() {
        loadBox.dispose();
        try {
            inStream.close();
        } catch (IOException ioe) {}
        loadBox  = null;
        inStream = null;
    }
    
    /*
     * Initialisiert den ViewerFrame. cores mu ein Vector von CoreStatus-
     * Objekten sein.
     */
    private void initFrame(GameInfo gameInfo, Vector cores) {
        
        /* Unter Windows die Hintergrundfarbe korrigieren. */
        if (System.getProperty("os.name").startsWith("Windows")) {
            setBackground(SystemColor.control);
        }

        /*
         * Links wird das displayPanel mit RAM und Coreliste angeordnet und
         * rechts das rightPanel mit Statusanzeige, Legende und Bedienelementen.
         */
        Panel displayPanel = new Panel(new BorderLayout());
        Panel rightPanel   = new Panel(new BorderLayout());

        ram = new RAMDisplay(gameInfo.getRAMSize(), cores);
        displayPanel.add(ram, "North");
        coreList = new ClientCoreList(10);
        coreList.setElements(cores);
        displayPanel.add(coreList, "Center");

        status = new StatusDisplay(gameInfo);
        //status.setVisible(StatusDisplay.INFO_TOTAL_ROUNDS, false);
        //status.setVisible(StatusDisplay.INFO_MAX_CYCLES, false);
        //status.setVisible(StatusDisplay.INFO_MAX_CORE_NUMBER, false);
        status.setVisible(StatusDisplay.INFO_MAX_CORE_SIZE, false);
        status.setVisible(StatusDisplay.INFO_MIN_DISTANCE, false);
        rightPanel.add(status, "North");
        rightPanel.add(makeLegend(cores), "Center");

        /* Das userPanel enthlt das toolPanel und das buttonPanel. */
        Panel userPanel   = new Panel(new BorderLayout());
        toolPanel         = new Panel();
        userPanel.add(toolPanel, "Center");
        Panel  buttonPanel = new Panel();
        Button exitButton  = new Button("Beenden");
        exitButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                broadcastClosingEvent();
            }
        });
        buttonPanel.add(exitButton);
        userPanel.add(buttonPanel, "South");
        rightPanel.add(userPanel, "South");
        
        add(displayPanel, "West");
        add(rightPanel, "Center");
    }
    
    /*
     * Alle WindowListener fr den ViewerFrame benachrichtigen, da das Fenster
     * geschlossen werden soll.
     */
    private void broadcastClosingEvent() {
        processWindowEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
    }

    /*
     * Erzeugt eine Legende fr die Farben der Cores.
     */
    private Component makeLegend(Vector cores) {
        ScrollPane scrollPane  = new ScrollPane();
        Panel      panel       = new Panel(new BorderLayout());
        Panel      legendPanel = new Panel(new GridLayout(0, 1));
        
        for (int i=0; i<cores.size(); i++) {
            CoreStatus core  = (CoreStatus) cores.elementAt(i);
            Label      label = new Label(core.author, Label.CENTER);
            label.setForeground(Color.black);
            label.setBackground(core.color);
            legendPanel.add(label);
        }
        panel.add(legendPanel, "North");        
        scrollPane.add(panel);
        return scrollPane;
    }
    
    /**
     * Zeigt das Anzeigefenster an.
     * @param listener  Wird als WindowListener fr das Fenster eingetragen.
     */
    public void show(WindowListener listener) {
        addWindowListener(listener);
        setVisible(true);
    }

    /**
     * Lt den Anwender in einem Dateidialog auswhlen, wo die History
     * gespeichert werden soll und fhrt die Speicherung durch. Der gesamte
     * Vorgang wird asynchron durchgefhrt, d.h. die Methode kehrt sofort
     * zurck. Nach dem Speichern wird savingFinished() aufgerufen.
     */
    public void saveHistory() {
        (new SaveThread()).start();
    }

    /*
     * Der Thread zum Speichern der History.
     */
    private class SaveThread extends Thread {

        /* Eine Kopie des zu speichernden GameInfo-Objekts. */
        private GameInfo gameInfo;

        /* Kopien der zu speichernden CoreStatus-Objekte. */
        private CoreStatus[] cores;

        /* Die StepHistory der zu speichernden Spielrunde.  */
        private StepHistory history;

        /*
         * Kopiert das GameInfo-Objekt und die CoreStatus-Objekte.
         */
        public SaveThread() {
            gameInfo = ViewerFrame.this.status.getGameInfo().createCopy();
            Vector origCores = ViewerFrame.this.coreList.getElements();
            cores = new CoreStatus[origCores.size()];
            for (int i=0; i<cores.length; i++) {
                cores[i] = (CoreStatus) ((CoreStatus) origCores.elementAt(i))
                                         .createCopy();
            }
            history = ViewerFrame.this.history;
        }

        /*
         * ffnet eine Dateidialog und speichert die History in der ausgewhlten
         * Datei.
         */
        public void run() {
            saveThreads++;
            FileDialog saveDialog = new FileDialog(new Frame(),
                    "History speichern", FileDialog.SAVE);
            saveDialog.show();
            
            if (saveDialog.getFile() != null) {
                ObjectOutputStream outStream = null;
                SaveBox            saveBox   = new SaveBox(saveDialog.getFile(),
                                                           null);
                try {
                    outStream = new ObjectOutputStream(new FileOutputStream
                                                       (saveDialog.getFile()));
                    saveBox.stream = outStream;
                    saveBox.setVisible(true);
                    gameInfo.setMaxCycles(gameInfo.getCycleNumber());
                    outStream.writeObject(gameInfo);
                    gameInfo = null;
                    for (int i=0; i<cores.length; i++) {
                        outStream.writeObject(cores[i]);
                    }
                    cores = null;
                    history.save(outStream);
                } catch (IOException ioe) {
                    if (!saveBox.getAborted()) {
                        (new MessageBox("History konnte nicht gespeichert "
                                        + "werden:\n" + ioe.getMessage(),
                                        "Fehler beim Speichern", saveBox,
                                        MessageBox.OK)).execute();
                    }
                }
                if (outStream != null) {
                    try {
                        outStream.close();
                    } catch (IOException ioe) {}
                }
                saveBox.dispose();
            }
            history = null;
            saveThreads--;
            savingFinished();
        }
    }
    
    /**
     * Diese Methode wird nach Beendigung des Speichervorgangs aufgerufen,
     * damit abgeleitete Klassen darauf reagieren knnen. In dieser Klasse
     * tut die Methode gar nichts.
     */
    protected void savingFinished() {}
    
    /**
     * Startet eine Spielrunde.
     * @param startPC  Die Startadressen aller beteiligten Cores.
     */
    public void startRound(int[] startPC) {
        playing = true;
        setCoresAlive();
        GameInfo gameInfo = status.getGameInfo();
        status.setActiveCores(gameInfo.getCoreNumber());
        status.incrementRound();
        status.setCycle(0);
        history = new StepHistory(ram, startPC, gameInfo.getMaxCycles());
    }

    /**
     * Fhrt nach dem Ende einer Spielrunde die Rundenauswertung durch.
     * @return  true, wenn das die letzte Runde war.
     */
    public boolean endRound() {
        Vector cores = coreList.getElements();
        for (int i=0; i<cores.size(); i++) {
            CoreStatus core = (CoreStatus) cores.elementAt(i);
            if (core.alive) {
                core.score++;
            }
        }
        coreList.refreshAll();
        playing = false;

        GameInfo gameInfo = status.getGameInfo();
        return gameInfo.getRoundNumber() == gameInfo.getTotalRounds();
    }

    /**
     * Spielt einen Zyklus aus der history vorwrts ab.
     * @return  Die aktuelle Zyklusnummer.
     */
    public int cycleForward() {
//        System.out.println("cycleForward()");
        status.incrementCycle();
        GameInfo gameInfo = status.getGameInfo();
        for (int i=0; i<gameInfo.getActiveCores(); i++) {

            if (history.stepForward()) {
                killCore(i);
                i--;
                status.setActiveCores((byte) (gameInfo.getActiveCores() - 1));
            }
        }
        return gameInfo.getCycleNumber();
    }

    /**
     * Spielt einen Zyklus aus der history rckwrts ab.
     * @return  Die aktuelle Zyklusnummer.
     */
    public int cycleBackward() { 
//        System.out.println("cycleBackward()");
        GameInfo gameInfo = status.getGameInfo();
        int      cycleNum = gameInfo.getCycleNumber() - 1;
        int      numCores = gameInfo.getActiveCores();
        status.setCycle(cycleNum);
        
//        byte prevCore = history.previousCore();
//        while (prevCore
        
        
        for (int i=0; i<numCores; i++) {
            int reviveNum = history.stepBackward();
            if (reviveNum != -1) {
//                System.out.println("core wiederbelebt");
                coreList.setState(reviveNum, true);
                numCores++;
                status.setActiveCores((byte) numCores);
            }
        }
        return cycleNum;
    }
    
    /*
     * Setzt den Status des (num+1)-ten noch aktiven Cores auf "tot".
     */
    private void killCore(int num) {
        Vector cores = coreList.getElements();
        int    index = -1;
        int    count = 0;
        do {
            index++;
            if (((CoreStatus) cores.elementAt(index)).alive) {
                count++;
            }
        } while (count <= num);
        coreList.setState(index, false);
    }

    /*
     * Setzt den Status aller Cores auf aktiv.
     */
    private  void setCoresAlive() {
        Vector cores = coreList.getElements();
        for (int i=0; i<cores.size(); i++) {
            ((CoreStatus) cores.elementAt(i)).alive = true;
        }
        coreList.refreshAll();
    }
    
    public static void main(String[] args) {
        ParamInfo param = new ParamInfo();
        param.setRAMSize(8192);
        Vector cores = new Vector();
        CoreInfo core = new CoreInfo();
        core.coreName = "Super Core";
        core.author = "Coremaster";
        core.size = 364;
        cores.addElement(core);
        core = new CoreInfo();
        core.coreName = "Raupe";
        core.author = "Anonymous";
        core.size = 6;
        cores.addElement(core);
        core = new CoreInfo();
        core.coreName = "Dabeiseinisalles";
        core.author = "Loser";
        core.size = 142;
        cores.addElement(core);
        for (int i=0; i<17; i++) {
            core = new CoreInfo();
            core.coreName = "TestCore";
            core.author = "TestAutor";
            core.size = 123;
            cores.addElement(core);
        }
        GameSetupInfo setup = new GameSetupInfo(param, cores, (byte) 0);
        ViewerFrame viewer = new ViewerFrame(setup);
        viewer.pack();
        viewer.show(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                System.exit(0);
            }
        });
    }
}

/**
 * Die abstrakte Klasse stellt ein Statusfenster mit einem "Abbrechen"-Knopf
 * dar. Abgeleitete Klassen mssen die abstrakte Funktion actionAborted()
 * berschreiben, um auf einen Abbruch zu reagieren.
 */
abstract class StatusBox extends Frame {

    private boolean aborted = false;

    /**
     * Initialisiert die StatusBox.
     * @param message  Die Nachricht, die in der StatusBox angezeigt werden
     *                 soll.
     * @param title    Der Titel der StatusBox.
     */
    public StatusBox(String message, String title) {
        super(title);
        
        /* Unter Windows die Hintergrundfarbe korrigieren. */
        if (System.getProperty("os.name").startsWith("Windows")) {
            setBackground(SystemColor.control);
        }       
        
        add(new Label(message, Label.CENTER), "Center");
        Panel  buttonPanel  = new Panel();
        Button cancelButton = new Button("Abbrechen");
        cancelButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                aborted = true;
                actionAborted();
                dispose();
            }
        });
        buttonPanel.add(cancelButton);
        add(buttonPanel, "South");

        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                aborted = true;
                actionAborted();
                dispose();
            }
        });
        pack();
    }

    /**
     * Reagiert auf einen Abbruch.
     */
    public abstract void actionAborted();
    
    /**
     * Liefert true, wenn abgebrochen wurde.
     */
    public boolean getAborted() {
        return aborted;
    }
}

/**
 * Die Klasse LoadBox stellt ein Statusfenster fr das Laden einer Historydatei
 * dar.
 */
class LoadBox extends StatusBox {

    /* Der InputStream, der bei einem Abbruch geschlossen wird. */
    InputStream stream;

    /**
     * Initialisiert die LoadBox.
     * @param fileName  Der Name der Historydatei.
     * @param stream    Der InputStream, aus dem die History gelesen wird. Er
     *                  wird bei einem Abbruch geschlossen.
     */
    public LoadBox(String fileName, InputStream stream) {
        super(fileName + " wird geladen...", "Laden");
        this.stream = stream;
    }

    /**
     * Schliet den InputStream.
     */
    public void actionAborted() {
        try {
            stream.close();
        } catch (IOException ioe) {}
    }
}

/**
 * Die Klasse SaveBox stellt ein Statusfenster fr das Speichern einer
 * Historydatei dar.
 */
class SaveBox extends StatusBox {

    /* Der OutputStream, der bei einem Abbruch geschlossen wird. */
    OutputStream stream;

    /**
     * Initialisiert die SaveBox.
     * @param fileName  Der Name der Historydatei.
     * @param stream    Der OutputStream, aus dem die History gelesen wird. Er
     *                  wird bei einem Abbruch geschlossen.
     */
    public SaveBox(String fileName, OutputStream stream) {
        super("History wird in " + fileName + " gespeichert...", "Speichern");
        this.stream = stream;
    }

    /**
     * Schliet den OutputStream.
     */
    public void actionAborted() {
        try {
            stream.close();
        } catch (IOException ioe) {}
    }
}
