/*
 * @(#)MetaServer.java    27.6.1999
 */


package corewar.metaserver;

import corewar.ParamDisplay;
import corewar.MessageBox;
import corewar.ToggleAction;
import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.io.IOException;
import java.util.Vector;

/**
 * Die Klasse MetaServer stellt das Hauptfenster des Verteilers dar und enthlt
 * Funktionen zum Verteilern der Clients auf die Server.
 *
 * @author  Ivo Kasiuk
 */
public class MetaServer extends Frame {

    /* Die Beschriftungen fr den "Sperren"-Button. */
    private static final String[] lockLabels = { "Sperren", "Freigeben" };
    
    /* Der ServerSocket fr sich anmeldende Server. */
    private MetaServerSSocket serverSSocket;

    /* Der ServerSocket fr sich anmeldende Clients. */
    private MetaClientSSocket clientSSocket;

    /* Die Liste mit den MetaServerSocket-Objekten. */
    private MetaServerList serverList;

    /* Die Liste mit den MetaClientSocket-Objekten. */
    private MetaClientList clientList;

    /*
     * Das Anzeigefeld fr die Parameter des in der Serverliste ausgewhlten
     *  Servers.
     */
    private ParamDisplay paramDisplay;

    /*
     * Die laufende Identifikationsnr. fr die Anmeldung der Clients bei
     * den Servern.
     */
    private byte clientServerID = 0;
    
    /**
     * Erzeugt das Verteilerfenster und richtet das MetaServerSSocket- und das
     * MetaClientSSocket-Objekt ein.
     */
    public MetaServer() {
        super("Verteiler");

        /* Unter Windows die Hintergrundfarbe korrigieren. */
        if (System.getProperty("os.name").startsWith("Windows")) {
            setBackground(SystemColor.control);
        }

        Panel listPanel = new Panel(new GridLayout(1, 2));
        clientList = new MetaClientList(20);
        serverList = new MetaServerList(20);
        serverList.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent event) {
                showSelectedServer();
            }
        });
        listPanel.add(clientList);
        listPanel.add(serverList);
        add(listPanel, "Center");

        Panel rightPanel = new Panel(new BorderLayout());
        paramDisplay = new ParamDisplay();
        rightPanel.add(paramDisplay, "North");

        GridBagLayout      gridBag     = new GridBagLayout();
        GridBagConstraints constraints = new GridBagConstraints();
        Panel              buttonPanel = new Panel(gridBag);
        constraints.fill      = GridBagConstraints.HORIZONTAL;
        constraints.weightx   = 1.0;
        constraints.gridwidth = GridBagConstraints.REMAINDER;
        
        Button lockButton = new Button();
        new ToggleAction(lockButton, lockLabels, false) {
            public void buttonToggled(boolean activated) {
                clientSSocket.setLocked(activated);
            }
        };
        constraints.insets = new Insets(3, 3, 3, 3);
        gridBag.setConstraints(lockButton, constraints);
        buttonPanel.add(lockButton);
        
        Button exitButton = new Button("Beenden");
        exitButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                exitMetaServer();
            }
        });
        constraints.insets = new Insets(5, 3, 3, 3);
        gridBag.setConstraints(exitButton, constraints);
        buttonPanel.add(exitButton);
        
        rightPanel.add(buttonPanel, "South");
        add(rightPanel, "East");
        
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                exitMetaServer();
            }
        });
        pack();

        try {
            serverSSocket = new MetaServerSSocket(this, false);
            clientSSocket = new MetaClientSSocket(this, false);
        } catch (IOException ioe) {
            String message = "Netzwerksocket fr "
                    + ((serverSSocket == null) ? "Server" : "Clients")
                    + " konnte nicht eingerichtet werden";
            (new MessageBox(message, "Fehler beim Starten", this,
                            MessageBox.OK)).execute();
            exitMetaServer();
        }
        serverSSocket.start();
        clientSSocket.start();
    }

    /*
     * Schliet alle offenen Netzwerkverbindungen und beendet den Verteiler.
     */
    private synchronized void exitMetaServer() {
        if (serverSSocket != null) {
            try {
                serverSSocket.close();
            } catch (IOException ioe) {}
        }
        if (clientSSocket != null) {
            try {
                clientSSocket.close();
            } catch (IOException ioe) {}
        }
        removeAllServers();
        removeAllClients();
        System.exit(0);
    }

    /*
     * Schliet die Verbindungen zu allen Servern und entfernt sie aus der
     * serverList.
     */
    private void removeAllServers() {
        Vector servers = serverList.getElements();
        while (servers.size() > 0) {
            MetaServerSocket server = (MetaServerSocket) servers.elementAt(0);
            serverList.removeElementAt(0);
            try {
                server.abortConnection();
            } catch (IOException ioe) {}
        }
    }

    /*
     * Schliet die Verbindungen zu allen Clients und entfernt sie aus der
     * clientList.
     */
    private void removeAllClients() {
        Vector clients = clientList.getElements();
        while (clients.size() > 0) {
            MetaClientSocket client = (MetaClientSocket) clients.elementAt(0);
            clientList.removeElementAt(0);
            try {
                client.abortConnection();
            } catch (IOException ioe) {}
        }
    }

    /*
     * Zeigt die Parameter des in der Serverliste ausgewhlten Servers im
     * Parameter-Anzeigefeld an.
     */
    private void showSelectedServer() {
        int index = serverList.getSelectedIndex();
        if (index < 0) {
            paramDisplay.setParamInfo(null);
        } else {
            MetaServerSocket server = (MetaServerSocket)
                    serverList.getElements().elementAt(index);
            paramDisplay.setParamInfo(server.getParamInfo());
        }
    }
    
    /**
     * Fgt einen Server zur Serverliste hinzu.
     * @param server  Der hinzuzufgende Server.
     */
    public synchronized void addServer(MetaServerSocket server) {
        serverList.addElement(server);
    }

    /**
     * Fgt einen Client zur Clientliste hinzu.
     * @param client  Der hinzuzufgende Client.
     */
    public synchronized void addClient(MetaClientSocket client) {
        clientList.addElement(client);
    }

    /**
     * Entfernt einen Server aus der Serverliste.
     * @param server  Der zu entfernende Server.
     */
    public synchronized void removeServer(MetaServerSocket server) {
        Vector servers = serverList.getElements();
        for (int i=0; i<servers.size(); i++) {
            if (servers.elementAt(i) == server) {
                serverList.removeElementAt(i);
                showSelectedServer();
                break;
            }
        }
    }

    /**
     * Entfernt einen Client aus der Clientliste.
     * @param client  Der zu entfernende Client.
     */
    public synchronized void removeClient(MetaClientSocket client) {
        Vector clients = clientList.getElements();
        for (int i=0; i<clients.size(); i++) {
            if (clients.elementAt(i) == client) {
                clientList.removeElementAt(i);
                break;
            }
        }
    }

    /**
     * Versucht einem Server Clients aus der Clientliste zuzuweisen.
     * @param server  Der server, dem Clients zugewiesen werden sollen.
     */
    public synchronized void distributeServer(MetaServerSocket server) {
        refreshServer(server);
        
        Vector clients = clientList.getElements();
        int    maxSize = server.getParamInfo().getMaxCoreSize();
        for (int i=0; (server.getNumCoresNeeded() > 0) && (i < clients.size());
                i++) {
            MetaClientSocket client = (MetaClientSocket) clients.elementAt(i);
            if (client.getCoreInfo().size <= maxSize) {
                try {
                    sendClientToServer(client, server);
                } catch (IOException ioe) {
                    break;
                }
                i--;
            }
        }
    }

    /**
     * Versucht einen Client einem Server aus der Serverliste zuzuordnen.
     * Wenn dies nicht gelingt, wird der Client int die Liste aufgenommen.
     * @param client  Der Client, dem ein Server zugeordnet werden soll.
     */
    public synchronized void distributeClient(MetaClientSocket client) {
        Vector servers  = serverList.getElements();
        int    coreSize = client.getCoreInfo().size;
        for (int i=0; i<servers.size(); i++) {
            MetaServerSocket server = (MetaServerSocket) servers.elementAt(i);
            if ((server.getNumCoresNeeded() > 0)
                    && (server.getParamInfo().getMaxCoreSize() >= coreSize)) {
                try {
                    sendClientToServer(client, server);
                    return;
                } catch (IOException ioe) {
                    i--;
                }
            }
        }
        addClient(client);
        client.start();
    }

    /*
     * Ordnet einen Client einem Server zu.
     */
    private void sendClientToServer(MetaClientSocket client,
            MetaServerSocket server) throws IOException {
        if (++clientServerID == 0) {
            clientServerID++;
        }
        server.thisIsYourClient(client, clientServerID);
        removeClient(client);
        refreshServer(server);
    }

    /*
     * Stellt die zu server gehrende Zeile der Serverliste neu dar.
     */
    private void refreshServer(MetaServerSocket server) {
        Vector servers = serverList.getElements();
        for (int i=0; i<servers.size(); i++) {
            if (servers.elementAt(i) == server) {
                serverList.refreshLine(i);
                break;
            }
        }
    }
    
    /**
     * Startet das Verteiler-Programm, indem das Hauptfenster erzeugt und
     * angezeigt wird.
     */
    public static void main(String[] args) {
        (new MetaServer()).setVisible(true);
    }
}
