/*
 * @(#)Interpreter.java   19.06.1999
 */
 
package corewar.server;


/**
 * Die Klasse fuehrt die Schritte einer
 * Runde durch.
 *
 * @author Burkhard Moser
 */
public class Interpreter
{
   /** ram  Enthaelt eine Referenz auf
    *       das Ram, den speicher, in dem sich die
    *       Cores befinden.*/
   Ram ram;
   

   /**
    * Initialisiert das Objekt.
    * 
    * @param ram  Referenz auf das Ram.
    */
   public Interpreter(Ram ram)
   {
      this.ram = ram;
   }


   /**
    * Fuehrt einen Befehl aus, indem an der
    * Stelle pc im Ram gelesen und der dort
    * stehende Befehl ausgefuehrt wird.
    * Die Resultate des Befehls werden in
    * einer StepInfo zurueckgeliefert.
    * Wurde ein ungueltiger Befehl ausgefuehrt,
    * enthaelt StepInfo.writeSize den Wert -1.
    * 
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   public StepInfo executeStep(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Je nach Wert an der Stelle pc im Ram
       * in die entsprechende Methode zur
       * Verarbeitung des mit dem Wert
       * codierten Assembler-Befehls verzweigen.*/
      switch(ram.read8(pc))          
      {
         case 1:
            stepInfo = Op1(pc);
            break;

         case 2:
            stepInfo = Op2(pc);
            break;
         
         case 3:
            stepInfo = Op3(pc);
            break;

         case 4:
            stepInfo = Op4(pc);
            break;
         
         case 5:
            stepInfo = Op5(pc);
            break;

         case 6:
            stepInfo = Op6(pc);
            break;
         
         case 7:
            stepInfo = Op7(pc);
            break;

         case 8:
            stepInfo = Op8(pc);
            break;
         
         case 9:
            stepInfo = Op9(pc);
            break;

         case 10:
            stepInfo = Op10(pc);
            break;
         
         case 11:
            stepInfo = Op11(pc);
            break;

         case 12:
            stepInfo = Op12(pc);
            break;
         
         case 13:
            stepInfo = Op13(pc);
            break;

         case 14:
            stepInfo = Op14(pc);
            break;
         
         case 15:
            stepInfo = Op15(pc);
            break;

         case 16:
            stepInfo = Op16(pc);
            break;
         
         case 17:
            stepInfo = Op17(pc);
            break;

         case 18:
            stepInfo = Op18(pc);
            break;
         
         case 19:
            stepInfo = Op19(pc);
            break;

         case 20:
            stepInfo = Op20(pc);
            break;
         
         /**Falls an der Stelle PC ein
          * ungueltiger Wert steht,
          * beliebige Werte in stepInfo
          * schreiben, aber writeSize
          * auf -1 setzen !!!*/
         default:                         
         {
            stepInfo.pc = pc;
            stepInfo.writeAddr = 0;
            stepInfo.writeSize = -1;
         }
      }

      return stepInfo;
   }


   /**
    * Die Methode fhrt den Assemblerbefehl
    * ADD A,B
    * aus: der Inhalt von Speicherstelle A
    * wird zum Inhalt von Speicherstelle B
    * addiert und das Ergebnis in A abgelegt.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op1(int pc)
   {
      StepInfo stepInfo = new StepInfo();
      
      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Den Inhalt des Ram an der Adresse A zum
       * Inhalt an der Adresse B addieren.*/
      int result = ram.read16(addressA) + ram.read16(addressB);     

      /**Das Ergebnis an die Speicherstellen A und A+1 schreiben,
       * dabei auf ein word casten, was einer Rechnung
       * modulo 65536 entspricht.*/
      ram.write16(addressA, result);         

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 2;

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * AND A,B
    * aus: die Inhalte von Speicherstelle A
    * und B werden logisch mit einem AND
    * verknuepft und das Ergebnis in A abgelegt.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op2(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Den Inhalt des Ram an den Adressen A und B mit AND verknuepfen.*/
      int result = ram.read8Signed(addressA) & ram.read8Signed(addressB);     

      /**Das Ergebnis an die Speicherstelle A schreiben,
       * dabei auf ein byte casten.*/
      ram.write8(addressA, result);         

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 1;

      return stepInfo;
   }


   /**
    * Die Methode fhrt den Assemblerbefehl
    * BOZ A(O)
    * aus: verzweige relativ zum aktuellen
    * Programmzaehler um O Woerter, falls
    * der Inhalt von A Null ist.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op3(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Wenn der Inhalt von Speicherstelle A Null ist,
       * verzweigen durch Berechnung eines neuen PC.*/
      if (ram.read8Signed(addressA) == 0)                   
      {
         int offset = ram.read8Signed(pc + 1);
         
         /**Neuen PC aus altem PC + Offset errechnen.*/
         int newPc = pc + offset;                       

         /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
         stepInfo.pc = ram.calcPc(newPc);
         stepInfo.writeAddr = 0;
         stepInfo.writeSize = 0;
      }
      
      /**Falls Inhalt von Speicherstelle A NICHT Null ist,
       * mit dem naechsten Befehl weitermachen.*/
      else                                               
      {
         /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
         stepInfo.pc = ram.calcPc(pc + 4);
         stepInfo.writeAddr = 0;
         stepInfo.writeSize = 0;
      }

      return stepInfo;
   }


   /**
    * Die Methode fhrt den Assemblerbefehl
    * JMN A,B
    * aus: springe zu Speicherstelle A,
    * falls der Inhalt der Speicherstelle B negativ ist.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op4(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Wenn der Inhalt der Speicherstelle B negativ ist,
       * setze den PC auf Speicherstelle A.*/
      if (ram.read8Signed(addressB) < 0)                  
      {
         /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/

         /**Zu der zuvor berechneten Adresse springen.*/
         stepInfo.pc = addressA;                         
         stepInfo.writeAddr = 0;
         stepInfo.writeSize = 0;
      }
      
      /**Falls Inhalt von Speicherstelle B NICHT negativ ist,
       * mit dem naechsten Befehl weitermachen.*/
      else                                               
      {
         /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
         stepInfo.pc = ram.calcPc(pc + 6);
         stepInfo.writeAddr = 0;
         stepInfo.writeSize = 0;
      }

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * JMP A
    * aus: springe zu Speicherstelle A.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op5(int pc)
   {
      StepInfo stepInfo = new StepInfo();
   
      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/

      /**Zu der zuvor berechneten Adresse springen.*/
      stepInfo.pc = addressA;                            
      stepInfo.writeAddr = 0;
      stepInfo.writeSize = 0;

      return stepInfo;
   }


   /**
    * Die Methode fhrt den Assemblerbefehl
    * JMP @A
    * aus: springe zu der Speicherstelle,
    * auf die Speicherstelle A zeigt.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op6(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int directAddressA = ram.calcPc(ram.read16(pc + 2));    

      /**Die Speicherstelle bestimmen, auf die der Inhalt der
       * Speicherstelle A zeigt und modulo ramSize rechnen.*/
      int indirectAddressA = ram.calcPc(ram.read16(directAddressA));   

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/

      /**Zu der zuvor berechneten Adresse springen.*/
      stepInfo.pc = indirectAddressA;           
      stepInfo.writeAddr = 0;
      stepInfo.writeSize = 0;

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * JMZ A,B
    * aus: springe zu Speicherstelle A,
    * falls Speicherstelle B den Wert Null enthaelt.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op7(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Wenn der Inhalt der Speicherstelle B Null ist,
       * setze den PC auf Speicherstelle A.*/
      if (ram.read8Signed(addressB) == 0)                  
      {
         /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/

         /**Zu der zuvor berechneten Adresse springen.*/
         stepInfo.pc = addressA;                         
         stepInfo.writeAddr = 0;
         stepInfo.writeSize = 0;
      }
      
      /**Falls Inhalt von Speicherstelle B NICHT Null ist,
       * mit dem naechsten Befehl weitermachen.*/
      else                                               
      {
      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = 0;
      stepInfo.writeSize = 0;
      }

      return stepInfo;
   }


   /**
    * Die Methode fhrt den Assemblerbefehl
    * MMR A,B(O)
    * aus: kopiere O+1 Bytes ab Adresse PC+A
    * an Adresse PC+B. Alle Werte fuer O
    * groesser als 5 sind unzulaessig.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op8(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Offset O bestimmen.*/
      int offset = ram.read8(pc + 1);                    

      /**Wenn Offset groesser als 5 ist, liegt ein ungueltiger Befehl vor,
       * der Core gilt als tot (=> writeSize = -1).*/
      if (offset > 5)                                     
      {
         /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
         stepInfo.pc = pc;
         stepInfo.writeAddr = 0;
         stepInfo.writeSize = -1;
      }
      
      /**Der Befehl ist gueltig, es werden O+1 Bytes kopiert.*/
      else                                               
      {
         /**O+1 Schleifendurchlaeufe.*/
         for (int i = 0; i <= offset; i++)               
         {
            /**Byte an der Stelle PC+A+i lesen.*/
            int byteToWrite = ram.read8Signed(ram.calcPc(pc + addressA +i));     

            /**Byte an die Stelle PC+B+i schreiben.*/
            ram.write8(ram.calcPc(pc + addressB + i), byteToWrite);               
         }

         /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
         stepInfo.pc = ram.calcPc(pc + 6);
         stepInfo.writeAddr = ram.calcPc(addressB + pc);
         stepInfo.writeSize = (byte) (offset +1);
      }

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * MOH A,B(O)
    * aus: kopiere das hoeherwertige Byte
    * des Inhalts der Adresse A an Adresse B+O.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op9(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Offset O bestimmen.*/
      int offset = ram.read8Signed(pc + 1);              

      /**Zieladresse bestimmen: Adresse B + offset.*/
      int target = addressB + offset;                    

      /**Das MSB von Adresse A(steht an addressA+1) lesen und
       * auf das MSB der Zieladresse schreiben (steht an target+1).*/
      ram.write8(target + 1, ram.read8Signed(addressA + 1));         

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = ram.calcPc(target + 1);
      stepInfo.writeSize = 1;

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * MOV A,B(O)
    * aus: kopiere den Inhalt der Adresse A
    * an Adresse B+O.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op10(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Offset O bestimmen.*/
      int offset = ram.read8Signed(pc + 1);              

      /**Zieladresse bestimmen: Adresse B + offset.*/
      int target = addressB + offset;                    

      /**Den zu kopierenden Wert A an die Zieladresse schreiben.*/
      ram.write16(target, ram.read16Signed(addressA));         

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = ram.calcPc(target);
      stepInfo.writeSize = 2;

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * MOV #A,B(O)
    * aus: kopiere den Wert A an Adresse B+O.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op11(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Offset O bestimmen.*/
      int offset = ram.read8Signed(pc + 1);              

      /**Zieladresse bestimmen: Adresse B + offset.*/
      int target = addressB + offset;                    

      /**Den zu kopierenden Wert A an die Zieladresse schreiben.*/
      ram.write8(target, ram.read8Signed(pc));         

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = ram.calcPc(target);
      stepInfo.writeSize = 1;

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * MOV @A,B(O)
    * aus: kopiere den Inhalt der Speicherstelle,
    * auf die der Inhalt der Speicherstelle A
    * zeigt, an Adresse B+O.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op12(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int directAddressA = ram.calcPc(ram.read16(pc + 2));    

      /**Die Speicherstelle bestimmen, auf die der Inhalt der
       * Speicherstelle A zeigt und modulo ramSize rechnen.*/
      int indirectAddressA = ram.calcPc(ram.read16(directAddressA));   

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Offset O bestimmen.*/
      int offset = ram.read8Signed(pc + 1);              

      /**Zieladresse bestimmen: Adresse B + offset.*/
      int target = addressB + offset;                    

      /**Das zu kopierende Byte von Adresse indirectAddressA
       * an die Zieladresse schreiben.*/
      ram.write8(target, ram.read8Signed(indirectAddressA));                                 

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = ram.calcPc(target);
      stepInfo.writeSize = 1;

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * MOV A,@B(O)
    * aus: kopiere den Inhalt der Speicherstelle A
    * an die Stelle, auf die der Inhalt von
    * Speicherstelle B, erhoeht um O, zeigt.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op13(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int directAddressB = ram.calcPc(ram.read16(pc + 4));    

      /**Die Speicherstelle bestimmen, auf die der Inhalt der
       * Speicherstelle B zeigt und modulo ramSize rechnen.*/
      int indirectAddressB = ram.calcPc(ram.read16(directAddressB));   

      /**Offset O bestimmen.*/
      int offset = ram.read8Signed(pc + 1);              

      /**Zieladresse bestimmen: indirectAddressB + offset.*/
      int target = indirectAddressB + offset;                    

      /**Das zu kopierende Byte von Adresse addressA an die Zieladresse schreiben.*/
      ram.write8(target, ram.read8Signed(addressA));                                 

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = ram.calcPc(target);
      stepInfo.writeSize = 1;

      return stepInfo;
      
   }


   /**Die Methode fhrt den Assemblerbefehl
    * MUL A,B
    * aus: multipliziere den Inhalt der
    * Speicherstelle A mit dem Inhalt der
    * Speicherstelle B. Die oberen 16 Bit
    * des Ergebnisses stehen an Speicherstelle B,
    * die unteren 16 Bit in Speicherstelle A.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op14(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Inhalt der Speicherstellen A und B auslesen und multiplizieren.*/
      int result = ram.read16(addressA) * ram.read16(addressB);    
                                                                   
      /**Untere 16 Bit des Ergebnisses in lower16 schreiben.*/
      short lower16 = (short) result;                              

      /**Obere 16 Bit des Ergebnisses in upper16 schreiben.*/
      short upper16 = (short) (result >> 16);                      

      /**Die niederwertigen 16 Bit in Speicherstelle A schreiben.*/
      ram.write16(addressA, lower16);                                

      /**Die hoeherwertigen 16 Bit in Speicherstelle B schreiben.*/
      ram.write16(addressB, upper16);                                

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 2;

      return stepInfo;
   }

   
   /**Die Methode fhrt den Assemblerbefehl
    * NEG A
    * aus: bilde das Zweier-Komplement
    * zum Inhalt der Speicherstelle A
    * und lege es in Speicherstelle A ab.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op15(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Inhalt der Speicherstelle A auslesen.*/
      int argument = ram.read8(addressA);               

      /**Das binaere Komplement zum Inhalt der Speicherstelle A bilden.*/
      argument = ~argument;                              

      /**Das Ergebnis wieder in die Speicherstelle A zurueckschreiben.*/
      ram.write8(addressA, argument);                   

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 4);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 1;

      return stepInfo;
   }

   
   /**Die Methode fhrt den Assemblerbefehl
    * SIC A
    * aus: speichere den PC in der Speicherstelle A.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op16(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    
      
      /**Den Wert des PCs in das Word ab Speicherstelle A schreiben.*/
      ram.write16(addressA, pc);                   

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 4);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 2;

      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * SLL A(O)
    * aus: verschiebe den Inhalt von
    * Speicherstelle A logisch um O Bits nach links.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op17(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Offset O bestimmen.*/
      int offset = ram.read8Signed(pc + 1);              

      /**Da Java bei Werten groesser 32 fuer den Datentyp int
       * modulo 32 rechnet, rechnen wir bei Werten groesser 8
       * fuer ein zu shiftendes Byte offset modulo 8.*/
      offset = offset % 8;                               

      /**Zu shiftendes Byte aus addressA auslesen.*/
      int byteToShift = ram.read8(addressA);             

      /**Byte shiften.*/
      byteToShift = byteToShift << offset;           

      /**Ergebnis wieder in die Speicherstelle A zurueckschreiben.*/
      ram.write8(addressA, byteToShift);                 

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 4);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 1;
      
      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * SRL A(O)
    * aus: verschiebe den Inhalt von
    * Speicherstelle A logisch um O Bits nach rechts.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op18(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Offset O bestimmen.*/
      int offset = ram.read8Signed(pc + 1);              

      /**Da Java bei Werten groesser 32 fuer den Datentyp int
       * modulo 32 rechnet, rechnen wir bei Werten groesser 8
       * fuer ein zu shiftendes Byte offset modulo 8.*/
      offset = offset % 8;                               

      /**Zu shiftendes Byte aus addressA auslesen, wobei eventuelle
       * vorzeichenbestimmende Einsen oberhalb des 8. Bits bereits
       * durch Nullen ersetzt werden.*/
      int byteToShift = ram.read8(addressA);             

      /**Byte shiften.*/
      byteToShift = byteToShift >> offset;           

      /**Ergebnis wieder in die Speicherstelle A zurueckschreiben.*/
      ram.write8(addressA, byteToShift);                 

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 4);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 1;
      
      return stepInfo;
   }

   
   /**Die Methode fhrt den Assemblerbefehl
    * SUB A,B
    * aus: subtrahiere den Inhalt der
    * Speicherstelle B von Speicherstelle A
    * und lege das Ergebnis in Speicherstelle A ab.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op19(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    

      /**Inhalt des words ab Speicherstelle B vom
       * Inhalt des words ab Speicherstelle A abziehen.*/
      int result = ram.read16(addressA) - ram.read16(addressB);    

      /**Ergebnis wieder in die Speicherstelle A zurueckschreiben.
       * Ein negatives Ergebnis muss nicht extra in ein positives
       * Ergebnis durch Addition von 65536 umgewandelt werden, da
       * jede geschriebene negative Zahl, die beim Lesen als
       * vorzeichenlose Zahl interpretiert wird, ohnehin im Wert
       * der negativen Zahl+65536 entspricht (2er-Komplement !).*/
      ram.write16(addressA, result);                              

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 2;
      
      return stepInfo;
   }


   /**Die Methode fhrt den Assemblerbefehl
    * XOR A,B
    * aus: die Inhalte der Speicherstellen A
    * und B werden logisch mit einem XOR
    * verknuepft. Das Ergebnis steht in Speicherstelle A.
    *
    * @param pc  Gibt die Stelle im Ram an, an der
    *            der Programmzaehler des Cores liegt,
    *            fuer den jetzt ein Befehl ausgefuehrt
    *            werden soll.
    * @return    Eine StepInfo mit den ausgefuehrten
    *            Befehl beschreibenden Parametern.
    */
   private StepInfo Op20(int pc)
   {
      StepInfo stepInfo = new StepInfo();

      /**Speicherstelle A bestimmen und modulo ramSize rechnen.*/
      int addressA = ram.calcPc(ram.read16(pc + 2));    

      /**Speicherstelle B bestimmen und modulo ramSize rechnen.*/
      int addressB = ram.calcPc(ram.read16(pc + 4));    
      
      /**Inhalt der Speicherstellen A und B mit XOR verknuepfen.*/
      int result = ram.read8(addressA) ^ ram.read8(addressB);    

      /**Ergebnis wieder in die Speicherstelle A zurueckschreiben.*/
      ram.write8(addressA, result);                              

      /**Die Werte in der StepInfo entsprechend dem ausgefuehrten Befehl setzen.*/
      stepInfo.pc = ram.calcPc(pc + 6);
      stepInfo.writeAddr = addressA;
      stepInfo.writeSize = 1;
      
      return stepInfo;
   }
}
