/*
 * @(#)NetManager.java    15.6.1999
 */
package corewar.server;

import corewar.CWProtocol;
import corewar.MessageBox;
import corewar.ParamInfo;
import corewar.CoreInfo;
import java.io.IOException;
import java.net.InetAddress;

/**
 * Die Klasse NetManager ist die Schnittstelle zwischen dem ServerFrame und den
 * Netzwerkobjekten und haelt Referenzen auf fast alle Objekte, die mit der
 * Netzwerkkommunikation zu tun haben. Hier werden die Daten, die von ServerFrame
 * kommen evtl. bearbeitet und an die entsprechenden Stellen weitergeleitet.
 * Umgekehrt werden auch Daten, die von den einzelnen Objekten kommen an den
 * ServerFrame oder die entsprechenden Stellen weitergeleitet.
 *
 * @author  Stephan Keitz
 */
public class NetManager implements CWProtocol{

  /** Zeigt an, ob die Clients sich direkt anmelden duerfen, oder vom Metaserver
    * zugewiesen werden */
  private boolean independentMode = false;
  /** Referenz auf den ListWatcher */
  private ListWatcher listWatcher = new ListWatcher(this);
  /** Referenz auf den ServerFrame */
  private ServerFrame serverFrame = null;
  /** Referenz auf den ServerMetaSocket */
  private ServerMetaSocket serverMetaSocket = null;
  /** Referenz auf den ServerClientSSocket */
  private ServerClientSSocket serverClientSSocket = null;
  /** Portnummer des ServerClientSSocket */
  private int portNumber = DEFAULT_SERVER_PORT;
  /** Referenz und anlegen eines neuen ClientCommunicators */
  private ClientCommunicator clientCommunicator = new ClientCommunicator(this);

  /**
   * Der Konstruktor merkt sich die Referenz auf den ServerFrame und versucht einen
   * ServerClientSSocket aufzubauen, wobei er mehrer Ports (MAX_SERVER_PORTS)
   * durchprobiert.
   */
  public NetManager(ServerFrame sf) {
    serverFrame = sf;
    boolean created = false;
    byte portCount = 0;
    while (!created && (portCount < MAX_SERVER_PORTS)) {
      created = true;
      try {
        serverClientSSocket = new ServerClientSSocket(this, DEFAULT_SERVER_PORT + portCount++);
        serverClientSSocket.start();
        portNumber = DEFAULT_SERVER_PORT + portCount - 1;
      }
      catch (IOException e) {
        created = false;
        if (portCount >= MAX_SERVER_PORTS) {
          MessageBox mb = new MessageBox(
            "Serverport konnte nicht erstellt werden !\n"+e.getMessage(),
            "NetManager Fehler !",serverFrame, MessageBox.OK);
          int result = mb.execute();
        }
      }
    }
  }

  /**
   * Oeffnet und schliesst den ServerClientSSocket. Wenn der ServerClientSSocket
   * geschlossen ist werden alle ankommenden Clients abgewiesen.
   * @param locked  Zustand des ServerClientSSocket (offen/geschlosen)
   */
  public void setServerClientSSocketLocked(boolean locked) {
    serverClientSSocket.setLocked(locked);
  }

  /**
   * Fuer den evtl. Fall das sich waehrend einer Simulation alle Clients
   * abmelden oder wegen Netzwerkproblemen wegfallen wird dies dem ServerFrame
   * gemeldet.
   */
  public void noMoreClients() {
    serverFrame.noMoreClients();
  }

  /**
   * Liefert die Portnummer des ServerClientSSockets zurueck.
   * @return  Die Portnummer
   */
  public int getPortNumber() {
    return portNumber;
  }

  /**
   * Gibt zurueck in welchem Modus sich der Server befindet.
   * @return  true, wenn nicht an den Metaserver angemeldet
   */
  public boolean getIndependentMode() {
    return independentMode;
    
  }

  /**
   * Legt fest in welchem Modus sich der Server befindet
   * @param independent  true, wenn nicht an den Metaserver angemeldet
   */
  public void setIndependentMode(boolean independent) {
    independentMode = independent;
    setServerClientSSocketLocked(!independent);
  }

  /**
   * Wenn ein Client sich erfolgreich angemeldet hat wird sein
   * ServerClientSocket im ClientCommunicator eingefuegt und dem
   * ServerFrame wird eine CoreInfo uebergeben, der diese dann in eine
   * Liste einfuegt.
   * @param scs  Der Socket des Client
   * @param ci  Die zu Uebergebene CoreInfo
   */
  public synchronized void addClient(ServerClientSocket scs, CoreInfo ci) {
    serverFrame.addCoreInfo(ci);
    clientCommunicator.addClient(scs);
  }

  /**
   * Stellt eine Verbindung zum Metaserver her.
   * @param metaServer  Adresse des Metaservers
   * @exception IOException  Wenn die Verbindung nicht zustande kommt
   */
  public void connectToMeta(InetAddress metaServer) throws IOException {
    serverMetaSocket = new ServerMetaSocket(this, metaServer);
    serverMetaSocket.start();
    setServerClientSSocketLocked(false);
  }
  
  /**
   * Schliesst die Verbindung zum Metaserver.
   */
  public void closeMetaConnect() {
    try {
      serverMetaSocket.abortConnection();
    }
    catch (IOException e) {}
    setServerClientSSocketLocked(true);
    serverMetaSocket = null;
    }

  /**
   * Meldet, wenn die Verbindung zum Metaserver abbricht.
   */
  public void metaConnectionClosed() {
    serverFrame.metaConnectionClosed();
    setServerClientSSocketLocked(true);
    serverMetaSocket = null;
  }

  /**
   * Fuegt einen Client in die Warteliste des ListWatchers ein.
   * @param id  Der einzufuegende Client
   */
  public void queueID(byte id) {
    listWatcher.queueID(id);
  }

  /**
   * Loescht einen erwarteten Client aus der Warteliste des ListWatchers und liefert
   * ein 'true' zurueck, wenn sich der Client noch in der Liste befand.
   * @param id  Der zu loeschende Client
   * @return    true, wenn der Client noch in der Liste war
   */
  public boolean deleteIDifpossible(byte id) {
    return listWatcher.deleteIDifpossible(id);
  }

  /**
   * Erfragt beim ServerFrame die Spielparameter.
   * @return  Die Spielparameter
   */ 
  public ParamInfo getParamInfo() {
    return serverFrame.getParamInfo();
  }

  /**
   * Liefert die Anzahl der Cores, die schon im CoreInfoVector sind.
   */
  public int getCoreInfoVectorSize() {
    return (serverFrame.getCoreInfoVector()).size();
  }

  /**
   * Gibt die Anzahl der noch gebrauchten Cores zurueck, die sich ermitteln
   * aus den bereits erhaltenen Cores und dan maximal zulaessigen Cores.
   * @return  Die Anzahl der noch gebrauchten Cores
   */
  public byte getCoresNeeded() {
    byte coresTotalNeed = (serverFrame.getParamInfo()).getCoreNumber();
    int coresInList    = getCoreInfoVectorSize();
    return (byte)(coresTotalNeed - coresInList);
  }

  /**
   * Schliesst die Verbindung zu allen Clients.
   */
  public void closeClients() {
    clientCommunicator.closeClients();
  }

  /**
   * Sendet Daten an ServerMateSocket und ClientCommunicator weiter
   * @param id    Typ der Daten
   * @param data  Die Daten selbst
   */
  public void sendData(byte id, Object data) throws IOException {
    if (id == CORES_NEEDED) serverMetaSocket.sendData(id, data);
    else clientCommunicator.sendData(id, data);
  }

  /**
   * Meldet dem ServerFrame, dass das Spiel gestartet werden kann.
   */
  public void startTournament() {
    serverFrame.startTournament();
  }
}
