/*
 * @(#)ParamModule.java    13.6.1999
 */


package corewar.server;

import corewar.GameInfo;
import corewar.ParamInfo;
import corewar.CoreInfo;
import corewar.MessageBox;

import java.awt.GridLayout;
import java.awt.Button;
import java.awt.Label;
import java.awt.TextField;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Dialog;
import java.awt.Frame;
import java.awt.Color;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import java.util.Vector;
import java.lang.NumberFormatException;
import java.lang.Exception;

/**
 * Die Klasse ParamModul enthlt die Parameter eines Servers, kann diese auch mit Hilfe einer
 * Dialogbox verndern, sowie laden und Speichern.
 *
 * @author  Carsten Buschmann
 */
class ParamModule extends GameInfo {
	
	  //Ober- und Untergrenzen der Enstellbereiche der Parameter
	  protected static final int MIN_ROUNDS = 1;
    protected static final int MAX_ROUNDS = 100;
    protected static final int MIN_CYCLES = 1;
    protected static final int MAX_CYCLES = 100000;
    protected static final int MAX_RAM_SIZE = 65536;
    protected static final int MIN_RAM_SIZE = 1024;
    protected static final int MAX_CORE_SIZE = 512;
    protected static final int MIN_CORE_SIZE = 4;
    protected static final int MAX_CORE_DISTANCE = 1024;
    protected static final int MIN_CORE_DISTANCE = 0;
    protected static final int MAX_PLAYERS = 20;
    protected static final int MIN_PLAYERS = 2;

    //Referenz auf den Serverframe
    private ServerFrame serverFrame;
	  //Rckgabewert der Dialogbox
	  private int result;
	
    //Referenz auf den Dialog zum Einstellen der Parameter
    private Dialog dialog;
    //Referenz auf die Felder zum Eingeben der Parameter
    private TextField roundTextField;
    private TextField cycleTextField;
    private TextField countTextField;
    private TextField sizeTextField;
    private TextField lengthTextField;
    private TextField distTextField;

    //Dateiname der Konfigurationsdatei
    public static String fileName = "wwcwserv.cfg";

    /**
     * Erzeugt ein neues ParamModule-Objekt und ldt dessen Einstellungen vom Festspeicher
     * @param sf Referenz auf das zugehrige ServerFrame-Objekt
     */
    public ParamModule(ServerFrame sf) {
      serverFrame = sf;
      load();
    }

    /**
     * Liefert die Anzahl der CoreInfo-Objekte in der Coreliste des Serverframes.
     * @return Anzahl der CoreInfo-Objekte in der Coreliste 
     */
    private byte coresInList() {
    	return (byte)(serverFrame.getCoreInfoVector()).size();    	
    }
    
    /**
     * Prft die Parameter auf Korrektheit und gibt ggf. Fehlermeldungen aus. Dabei wird geprft,
     * ob die zulssigen Grenzen fr die Parameter eigehalten werden, ob sich die Parameter 
     * mit evtl. bereits in der Coreliste des Servers befindlichen Cores vertragen und ob sie evtl.
     * in sich widersprchlich sind. 
     * @return true, wenn die Parameter korrekt sind, sonst false
     * @param params die zu berprfenden Parameter
     */
		private boolean correctParams(ParamInfo params) {
		  Vector v = serverFrame.getCoreInfoVector();
		  //boolean isCorrect;
		  
		  if (!( (params.getTotalRounds() >= MIN_ROUNDS) && (params.getTotalRounds() <= MAX_ROUNDS) )) {
		             Message( "Fehler", "Die Rundenzahl liegt auerhalb des erlaubten Bereiches (siehe Klammern).", serverFrame );
		  	         return false;
		  }
			if (!( (params.getMaxCycles() >= MIN_CYCLES) && (params.getMaxCycles() <= MAX_CYCLES) )) {
		             Message( "Fehler", "Die Zyklenzahl liegt auerhalb des erlaubten Bereiches (siehe Klammern).", serverFrame );
		  	         return false;
		  }
			if (!( (params.getCoreNumber() >= MIN_PLAYERS) && (params.getCoreNumber() <= MAX_PLAYERS) ) ) {
		             Message( "Fehler", "Die Spielerzahl liegt auerhalb des erlaubten Bereiches (siehe Klammern)." + Byte.toString(params.getCoreNumber()), serverFrame );
		  	         return false;
		  }
			if (!( (params.getRAMSize() >= MIN_RAM_SIZE) && (params.getRAMSize() <= MAX_RAM_SIZE) )) {
		             Message( "Fehler", "Die Speichergre liegt auerhalb des erlaubten Bereiches (siehe Klammern).", serverFrame );
		  	         return false;
		  }
			if (!( (params.getMaxCoreSize() >= MIN_CORE_SIZE) && (params.getMaxCoreSize() <= MAX_CORE_SIZE) )) {
		             Message( "Fehler", "Die max. Coregre liegt auerhalb des erlaubten Bereiches (siehe Klammern).", serverFrame );
		  	         return false;
		  }
		  if (!( (params.getMinDistance() >= MIN_CORE_DISTANCE) && (params.getMinDistance() <= MAX_CORE_DISTANCE) ) ) {
		             Message( "Fehler", "Der min. Abstand der Cores liegt auerhalb des erlaubten Bereiches (siehe Klammern).", serverFrame );
		  	         return false;
		  }
		  	    
		  if ( params.getCoreNumber() < coresInList() ) {
         Message( "Fehler", "Es sind bereits mehr Cores ("+ Byte.toString(coresInList()) +" Stck) in der Liste, als sie zulassen wollen. ", serverFrame );
         return false;
		  }
		  if ( params.getMaxCoreSize() < largestCoreInList() ) {
         Message( "Fehler", "Es ist bereits ein Core in der Liste, der grer ist ( " + Short.toString(largestCoreInList()) + " Byte), als sie zulassen wollen.", serverFrame );
         return false;
		  }
		  if ( params.getMaxCoreSize() * params.getCoreNumber() + params.getMinDistance() * ( params.getCoreNumber() - 1 ) > params.getRAMSize() ) {
         Message( "Fehler", "Die von Ihnen angegebenen Kombination von Speichergre, min. Coreabstand und max. Coregre ist in sich widersprchlich und daher unmglich.", serverFrame );
         return false;
  	  }
		  return true;	       	
  	}

    /**
     * Liefert die aktuellen Parameter
     * @return  ParamInfo-Objekt mit den aktuellen Parametern
     */
    public ParamInfo getParamInfo() {
    	
    	 ParamInfo p = new ParamInfo();
       p.setTotalRounds(getTotalRounds());
       p.setMaxCycles(getMaxCycles());
       p.setRAMSize(getRAMSize());
       p.setMaxCoreSize(getMaxCoreSize());
       p.setCoreNumber(getCoreNumber());
       p.setMinDistance(getMinDistance());
          
       return p;       
    }

    /**
     * Ermittelt die Gre des grten in der Corelist des ServerFrame befindlichen Cores
     * @return die Lnge des grten Cores 
     */
    private short largestCoreInList() {
    	Vector v = null;
    	if (serverFrame != null) v = serverFrame.getCoreInfoVector();
    	short max = 0;
    	if (v != null) {
    	  CoreInfo ci;
    	  for (int i=0; i < coresInList(); i++ ) {
    		  ci = (CoreInfo)v.elementAt(i);
    		  max = (short)Math.max( max, ci.size );
      	}
    	}
    	return max;
    }
     
    /**
     * Ldt die Parameter vom Festspeicher
     */
	  private void load() {
        try {
          FileInputStream fis = new FileInputStream(fileName);
          ObjectInputStream oip = new ObjectInputStream(fis);
          ParamInfo pi = (ParamInfo)oip.readObject();
          fis.close();
          setParamInfo( pi );
        }
        catch( Exception e ) {Message( "Fehler", "Parameter konnten nicht gelesen werden.", serverFrame ); }
	  }

    /**
     * Gibt eine Mitteilung in einer Dialogbox aus.
     * @param Title Titel der Dialogbox
     * @param Text Text der Mitteilung
     * @param parent Parent-Fenster der Dialogbox
     */
    private void Message(String Title, String Text, Frame parent) {
   	   MessageBox m = new MessageBox( Text, Title, parent, MessageBox.OK );
       m.setBackground(Color.lightGray);
   	   int i = m.execute();
    }

    /**
     * ffnet eine Dialogbox zum Einstellen der Parameter. Wird "Ok" gedrckt, werden die 
     * Parameter auf Korrektheit berprft und ggf. gesetzt.
     */
    public void paramDialog() {
        
        dialog = new Dialog( serverFrame, "Parameter einstellen", true);
        dialog.setLayout( new GridLayout(7,2) );
        dialog.setResizable( false );
        
        Label label = new Label( "Rundenzahl (1-100):" );
        dialog.add( label );
        roundTextField = new TextField(Byte.toString(getTotalRounds()));
        dialog.add( roundTextField );
        
        dialog.add( new Label( "Max. Zyklenzahl (1-100.000):" ) );
        cycleTextField = new TextField(Integer.toString(getMaxCycles()));
        dialog.add( cycleTextField );
        
        dialog.add( new Label( "Anzahl der Cores (2-20):" ) );
        countTextField = new TextField(Byte.toString(getCoreNumber()));
        dialog.add( countTextField );
        
        dialog.add( new Label( "Speichergre (1024-65536):" ) );
        sizeTextField = new TextField(Integer.toString(getRAMSize()));
        dialog.add( sizeTextField );
        
        dialog.add( new Label( "Max. Corelnge (4-512):" ) );
        lengthTextField = new TextField(Short.toString(getMaxCoreSize()));
        dialog.add( lengthTextField );
        
        dialog.add( label = new Label( "Min. Coreabstand (1-1024):" ) );
        distTextField = new TextField(Short.toString(getMinDistance()));
        dialog.add( distTextField );
        
        Button button = new Button( "Ok" );
        button.addActionListener(new OkListener());
        dialog.add(button);
        	
        button = new Button( "Abbrechen" );
        button.addActionListener(new CancelListener());
        dialog.add(button);
        	
        dialog.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                result = 0;
                event.getWindow().setVisible(false);
            }
        });
          
        dialog.pack();
        dialog.show();    	
    }
    
    /**
     * Speichert die Parameter auf Festspeicher. Dazu mu zunchst der Name des voreingestellten
     * Metaservers gelesen werden. Dann wird die Datei zum Schreiben geffnet und Parameter und
     * Metaservername geschrieben.
     */
    private void save() {
        String serverName = null;
        //Servernamen lesen
        try {
          FileInputStream fis = new FileInputStream(fileName);
          ObjectInputStream oip = new ObjectInputStream(fis);
          ParamInfo pi = (ParamInfo)oip.readObject();
          serverName = (String)oip.readObject();
          fis.close();
        }
        catch( Exception e ) {Message( "Fehler", "Parameter konnten nicht richtig gespeichert werden.", serverFrame ); }
        //Parameter und Servernamen schreiben.
        try {
          FileOutputStream fos = new FileOutputStream(fileName);
          ObjectOutputStream oop = new ObjectOutputStream(fos);
          oop.writeObject(getParamInfo());
          oop.writeObject(serverName);
          fos.close();
        }
        catch( Exception e ) {Message( "Fehler", "Parameter konnten nicht gespeichert werden.", serverFrame ); }
    }

    /**
     * Setzt die Parameter neu.
     * @param p die zu setzenden Parameter
     */
    private void setParamInfo(ParamInfo p) {
    	setTotalRounds( p.getTotalRounds());
    	setMaxCycles( p.getMaxCycles());
    	setRAMSize( p.getRAMSize());
    	setMaxCoreSize( p.getMaxCoreSize());
    	setCoreNumber( p.getCoreNumber());
    	setMinDistance( p.getMinDistance());
    }


    /**
     * Listerner des "Ok" Buttons des Parameter-Dialoges: berprft die einstellten
     * Parameter durch Aufruf von correctParams() und setzt diese ggf.
     */
     class OkListener implements ActionListener {
          
          public void actionPerformed(ActionEvent ae)
          {
          	 boolean error = false;
             ParamInfo p = new ParamInfo();
             try {
               p.setTotalRounds( Byte.parseByte(roundTextField.getText()));
               p.setMaxCycles( Integer.parseInt(cycleTextField.getText()));
               p.setRAMSize( Integer.parseInt(sizeTextField.getText()));
               p.setMaxCoreSize( Short.parseShort(lengthTextField.getText()));
               p.setCoreNumber( Byte.parseByte(countTextField.getText()));
               p.setMinDistance( Short.parseShort(distTextField.getText()));
               
             }
             catch( NumberFormatException e ) { 
             	  Message("Fehler","Bitte tragen Sie in alle Felder gltige Ziffern ein.", serverFrame );
             	  error = true;
             } 
             if ( !error && correctParams(p) ) {
             	 setParamInfo(p);
             	 save();
               dialog.setVisible(false);
             	 (serverFrame.getStatusDisplay()).setParamInfo(p);
             }
          }
     }

    /**
     * Listerner des "Abbrechen" Buttons des Parameter-Dialoges: schiet diesen, ohne die nderungen
     * zu bernehmen.
     */
     class CancelListener implements ActionListener {
           
          public void actionPerformed(ActionEvent ae)
          {
                dialog.setVisible(false);
          }
     }


}