/*
 * @(#)Ram.java   19.06.1999
 */
 
package corewar.server;


/**
 * Die Klasse dient der Simulation des Hauptspeichers.
 * Sie haelt die Daten des Speichers und kann Lese-
 * und Schreibzugriffe durchfuehren.
 *
 * @author Burkhard Moser
 */
public class Ram
{
   /** size  Enthaelt die Groesse des
    *        Speichers in Byte.*/
   private int size;

   /** virtualRam  Das Array stellt den Speicher dar,
    *              in den die Cores geschrieben werden.
    *              Es umfasst soviele bytes, wie in size festgelegt.*/
   private byte[] virtualRam;


   /**
    * Erzeugt ein Ram-Objekt.
    * 
    * @param size  Groesse des Ram in Byte.
    */
   public Ram(int size)
   {
      this.size = size;
      this.virtualRam = new byte[size];
   }


   /**
    * Liest die Speicherstelle pc aus und liefert sie zurueck.
    * Dabei wird der Wert als vorzeichenlose Zahl interpretiert.
    * 
    * @param pc  Stelle im Ram, an der gelesen werden soll.
    * @return    Der Inhalt der Speicherstelle pc,
    *            interpretiert als vorzeichenlose Zahl.
    */
   public int read8(int pc)
   {
      int data = virtualRam[calcPc(pc)];

      /**Den Wert von virtualRam an der Stelle pc, interpretiert als
       * vorzeichenlose Zahl, zurueckgeben. Das Vorzeichen wird durch
       * UND-Verknuepfung mit der Bitfolge 00...0011111111 neutralisiert.*/
      return (data & 255);                

/*    try
      {
         byte data = virtualRam[pc];
         return data;
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
         pc = calcPc(pc);
         read8(pc);
      }
*/
   }


   /**
    * Liest die Speicherstelle pc aus und liefert sie zurueck.
    * Dabei wird der Wert als vorzeichenbehaftete Zahl interpretiert.
    * 
    * @param pc  Stelle im Ram, an der gelesen werden soll.
    * @return    Der Inhalt der Speicherstelle pc,
    *            interpretiert als vorzeichenbehaftete Zahl.
    */
   public int read8Signed(int pc)
   {
      int data = virtualRam[calcPc(pc)];

      /**Den Wert von virtualRam an der Stelle pc, interpretiert als
       * vorzeichenbehaftete Zahl, zurueckgeben.*/
      return data;                
   }


   /**
    * Schreibt an die Speicherstelle pc die
    * unteren 8 Bit des Integer-Wertes data.

    * @param pc    Stelle im Ram, an die geschrieben werden soll.
    * @param data  Integer-Zahl, die den an die Stelle pc
    *              zu schreibenden Wert enthaelt.
    */
   public void write8(int pc, int data)
   {
      virtualRam[calcPc(pc)] = (byte) data; 
   }
   
   
   /**
    * Liest die Speicherstellen pc und pc+1 aus und liefert sie zurueck.
    * Dabei wird der Wert als vorzeichenlose Zahl interpretiert.
    * 
    * @param pc  Stelle im Ram, ab der gelesen werden soll.
    * @return    Der Wert des Words in den Speicherstellen
    *            pc und pc+1, interpretiert als vorzeichenlose
    *            Zahl, wobei das LSB in der Speicherstelle pc
    *            und das MSB in der Speicherstelle pc+1 steht.
    */
   public int read16(int pc)
   {
      /**Byte an der Stelle pc auslesen. Es entspricht dem LSB des Words.*/
      int lsb = virtualRam[calcPc(pc)];   

      /**Byte an der Stelle pc+1 auslesen. Es entspricht dem LSB des Words.*/
      int msb = virtualRam[calcPc(pc + 1)];   

      /**In LSB und MSB werden alle 1-bits oberhalb des 8. Bits durch Nullen
       * ersetzt, das MSB anschliessend um 8 Stellen nach links geshiftet
       * und mit dem LSB verODERt*/
      int data = (((msb & 255) << 8) | (lsb & 255));
      return data;
   }


   /**
    * Liest die Speicherstellen pc und pc+1 aus und liefert sie zurueck.
    * Dabei wird der Wert als vorzeichenbehaftete Zahl interpretiert.
    * 
    * @param pc  Stelle im Ram, ab der gelesen werden soll.
    * @return    Der Wert des Words in den Speicherstellen
    *            pc und pc+1, interpretiert als vorzeichenbehaftete
    *            Zahl, wobei das LSB in der Speicherstelle pc
    *            und das MSB in der Speicherstelle pc+1 steht.
    */
   public int read16Signed(int pc)
   {
      /**Byte an der Stelle pc auslesen. Es entspricht dem LSB des Words.*/
      int lsb = virtualRam[calcPc(pc)];   

      /**Byte an der Stelle pc+1 auslesen. Es entspricht dem LSB des Words.*/
      int msb = virtualRam[calcPc(pc + 1)];   

      /**Im LSB werden alle 1-bits oberhalb des 8. Bits durch Nullen ersetzt,
       * dem MSB bleiben alle 1-bits erhalten.
       * Das MSB wird um 8 Stellen nach links geshiftet
       * und mit dem LSB verODERt*/
      int data = ((msb << 8) | (lsb & 255));
      return data;
   }


   /**
    * Schreibt an die Speicherstelle pc die unteren 8 Bit
    * des Integer-Wertes data und an die Stelle pc+1 die
    * Bits 9 bis 16 von data.
    *
    * @param pc    Stelle im Ram, ab der geschrieben werden soll.
    * @param data  Integer-Zahl, die den an die Stelle pc
    *              zu schreibenden Wert enthaelt.
    */
   public void write16(int pc, int data)
   {
      byte lsb = (byte) data;
      byte msb = (byte) (data >> 8);

      virtualRam[calcPc(pc)] = lsb;
      virtualRam[calcPc(pc + 1)] = msb;
   }


   /**
    * Wandelt einen pc, dessen Wert groesser ist als das Ram, durch Rechnung
    * modulo Ramgroesse in einen gueltigen pc um, was Ausdruck der ringfoermigen
    * Struktur des Ram ist.
    *
    * @param pc  Wert des Programmzaehlers, der in einen
    *            gueltigen Wert umgewandelt werden soll.
    * @return    Wert des Programmzaehlers nach Umwandlung
    *            in einen gueltigen Wert.
    */
   public int calcPc(int pc)
   {
      /**Falls der pc negativ ist, vorher in einen positiven pc umwandeln.*/
      if (pc < 0)
      {
         pc = makePcPositive(pc);  
      }

      int newPc = pc % size;
      return newPc;
   }


   /**
    * Wandelt einen negativen pc durch Addition der Groesse des ram
    * in einen positiven pc um, was Ausdruck der ringfoermigen
    * Struktur des Ram ist.
    *
    * @param pc  Wert des Programmzaehlers, der in einen
    *            positiven Wert umgewandelt werden soll.
    * @return    Wert des Programmzaehlers nach Umwandlung
    *            in einen positiven Wert.
    */
   public int makePcPositive(int pc)
   {
      while(pc < 0)
      {
         pc = pc + size;
      }

      return pc;
   }
}

