/*
 * @(#)Scheduler.java   19.06.1999
 */
 
package corewar.server;

import corewar.CWProtocol;
import java.util.Vector;


/**
 * Die Klasse fuehrt die Zyklen einer Runde durch.
 *
 * @author Burkhard Moser
 */
public class Scheduler implements CWProtocol
{
   /** simulator  Referenz auf den Scheduler erzeugenden Simulator.*/
   private Simulator simulator;

   /** interpreter  Referenz auf den vom Scheduler erzeugten Interpreter.*/
   private Interpreter interpreter;

   /** actualCycle  Referenz auf die vom Scheduler erzeugte CycleInfo.*/
   private CycleInfo actualCycle;

   /** pcs  Referenz auf das die Programmzaehler enthaltende Array.*/
   private int[] pcs;

   /** stopped  Boolescher Wert, der angibt, ob das Turnier von aussen
    *           gestoppt wurde.*/
   private boolean stopped = true;


   /**
    * Initialisiert das Scheduler-Objekt und erzeugt einen Interpreter
    * zum Ausfuehren einzelner Schritte der Cores.
    * 
    * @param ram  Referenz auf das zu verwendende Ram,
    *             den eigentlichen Speicher, in dem sich
    *             die Cores befinden.
    * @param pcs  Array, in dem die Programmanfaenge der
    *             einzelnen Cores stehen.
    * @param simulator  Referenz auf den Simulator, der
    *                   den Scheduler erzeugt hat.
    */
   public Scheduler(Ram ram, int[] pcs, Simulator simulator)
   {
      this.pcs = pcs;
      this.simulator = simulator;
      this.interpreter = new Interpreter(ram);

      /**Die Arrays in CycleInfo so initialisieren, dass
       * sie Platz haben fuer alle Daten.*/
      actualCycle = new CycleInfo(pcs.length);     
   }


   /**
    * Fuehrt einen Zyklus durch, indem es fuer jeden im Spiel
    * befindlichen Core einen Befehl durch den Interpreter
    * ausfuehren laesst. Am Ende wird eine CycleInfo an den
    * Simulator versandt. Liefert false zurueck, wenn der
    * Zyklus durch eine Aufruf von stop() beendet wurde,
    * ansonsten true.
    * 
    * @return    Boolescher Wert, der angibt, ob das Turnier von aussen
    *            gestoppt wurde.
    */
   public boolean runCycle()
   {
      StepInfo stepInfo;

      /**Fuer jeden PC, also letztlich jeden Core, einen Befehl
       * ausfuehren, solange nicht Stop gedrueckt wurde.*/
      for (int i = 0; (i < pcs.length) && stopped; i++)       
      {
         stepInfo = interpreter.executeStep(pcs[i]);

         actualCycle.pc[i] = (short) stepInfo.pc;
         actualCycle.writeAddr[i] = (short) stepInfo.writeAddr;
         actualCycle.writeSize[i] = stepInfo.writeSize;

         /**Wenn ein Core gestorben ist, seinen PC
          * auf einen ungueltigen Wert setzen.*/
         if (stepInfo.writeSize == -1)                
         {
            pcs[i] = -1;
         }

         /**Lebt der Core noch, seinen PC aktualisieren.*/
         else
         {
            pcs[i] = stepInfo.pc;                     
         }
      }

      /**Nur wenn stopped noch true ist, also NICHT Stop
       * gedrueckt wurde, wird eine CycleInfo versandt
       * (andernfalls waere sie nicht vollstaendig).*/
      if(stopped)                                  
      {
         sendData(CYCLE_INFO, actualCycle);
      }
      
      /**Ist stopped false, wurde also Stop gedrueckt,
       * wird STOP an die Clients versandt.*/
      else                                         
      {
         sendData(STOP, null);
      }
      
      /**Die PCs mit ungueltigen Werten aus dem Array pcs aussortieren.*/
      renewPcs();                               

      return stopped;
   }


   /**
    * Stoppt den laufenden Zyklus, indem die Instanzvariable
    * stopped auf false gesetzt wird.
    */
   public void stop()
   {
      stopped = false; 
   }


   /**
    * Versendet Daten ueber den abgelaufenen Spielzyklus an den Simulator.
    * 
    * @param id    Zahl, die codiert um welche Art Information
    *              es sich bei dem folgenden Object data handelt.
    * @param data  Object, das die eigentlichen zu
    *              uebertragenden Daten enthaelt.
    */
   private void sendData(byte id, Object data)
   {
      simulator.sendData(id, data);  
   }


   /**
    * Sortiert die PCs von gestorbenen Cores aus dem Array
    * der Programmzaehler aus, damit in der naechsten Runde
    * vom Interpreter keine Anweisungen dieser Cores mehr
    * ausgefuehrt werden.
    */
   private void renewPcs()
   {
      int validPcs = 0;
      
      /**Jeden PC im Array betrachten.*/
      for (int i = 0; i < pcs.length; i++)         
      {
         /**Die PCs zaehlen, die gueltig sind, also ungleich -1.*/
         if (pcs[i] != -1)                         
         {
            validPcs++;
         }
      }

      /**Ein neues int-Array, dessen Groesse der Anzahl gueltiger PCs entspricht.*/
      int[] newPcs = new int[validPcs];
      
      /**Da jetzt weniger Cores im Spiel sind (genau validPcs Stck),
       * wird eine neue Cycle Info mit kleineren Arrays erzeugt und
       * der Instanzvariablen actualCycle zugewiesen.*/
      actualCycle = new CycleInfo(validPcs);

      int counter = 0;

      /**Jeden PC im Array betrachten.*/
      for (int i = 0; i < pcs.length; i++)         
      {
         /**Die gueltigen PCs in das Array newPcs schreiben.*/
         if (pcs[i] != -1)                         
         {
            newPcs[counter] = pcs[i];
            counter++;
         }
      }

      /**pcs durch newPcs, das Array, das nur noch gueltige
       * PCs enthaelt, ersetzen.*/
      pcs = newPcs;                                

      /**Die Instanzvariable continue_flag der Klasse Simulator
       * wird auf false gesetzt, wenn nur noch ein Core lebt,
       * oder alle Cores tot sind.*/
      if (pcs.length < 2)
      {
         simulator.set_continue(false);
      }
   }
}
