/*
 * @(#)Simulator.java   19.06.1999
 */
 
package corewar.server;

import corewar.CWProtocol;
import corewar.ParamInfo;
import corewar.CoreInfo;
import java.util.Vector;
import java.lang.Class;
import java.lang.Math;
import java.lang.Integer;
import java.lang.String;
import java.lang.Object;


/**
 * Die Klasse erzeugt ein Ram-Objekt, plaziert
 * die am Spiel teilnehmenden Cores per Zufall
 * darin und fuehrt schliesslich eine Runde durch.
 *
 * @author Burkhard Moser
 */
public class Simulator implements CWProtocol
{
   /** tournament  Referenz auf das den Simulator erzeugende Tournament-Objekt.*/
   private Tournament tournament;

   /** scheduler  Referenz auf den vom Simulator erzeugten Scheduler.*/
   private Scheduler scheduler;
   
   /** ram  Referenz auf das Ram-Objekt, in dem die Simulation ablaeuft.*/
   private Ram ram;
   
   /** ramSize  Speichert die Groesse des Ram in Bytes.*/
   private int ramSize;
   
   /** pcs  Array, das die Programmzaehler der Cores im Ram enthaelt.*/
   private int[] pcs = null;
   
   /** mappedCoreInfoVector  Clone des CoreInfoVector, in dem die Cores
    *  allerdings zufaellig umsortiert wurden.*/
   private Vector mappedCoreInfoVector;
   
   /** continue_flag  Boolesche Variable, die true ist, solange noch lebendige
    *  Cores vorhanden sind und false, sobald alle Cores tot sind.*/
   private boolean continue_flag = true;


   /**
    * Initialisiert das Simulator-Objekt, plaziert die Cores unter 
    * Wahrung der in den Spielparametern eingestellten Mindestabstaende
    * im Ram, versendet eine RamInfo mit den Positionen der Cores im Ram
    * an ServerFrame und erzeugt einen Scheduler.
    * 
    * @param tournament      Referenz auf das den Simulator erzeugende
    *                        Tournament-Objekt
    * @param coreInfoVector  Vector, der die Informationen ueber
    *                        die in das arm zu schreibenden Cores enthaelt.
    * @param params          ParamInfo-Objekt, das die eingestellten
    *                        Spielparameter enthaelt
    */
   public Simulator(Tournament tournament, Vector coreInfoVector, ParamInfo params)
   {
      this.tournament = tournament;
      this.ramSize = params.getRAMSize();

      short minDistance = params.getMinDistance();
      initRam();

      /**Schreibt die Cores an Zufallspositionen in das Ram,
       * nachdem zuvor deren Reihenfolge im CoreInfoVector
       * zufaellig umsortiert wurde.*/
      placeCores(minDistance, coreInfoVector);

      RamInfo ramInfo = new RamInfo();
      
      /**Die Reihenfolge der PCs wieder so waehlen, wie die
       * Reihenfolge der zugehoerigen Cores im urspruenglichen
       * CoreInfoVector war.*/
      int[] remappedPcs = remapPcs();

      /**Die wieder in die urspruengliche Reihenfolge gebrachten PCs
       * in das RamInfo-Objekt schreiben.*/
      ramInfo.addr = remappedPcs;        
      sendData(RAM_INFO, ramInfo);

      /**Ein Scheduler-Objekt erzeugen an Hand eines Arrays
       * von Programmzaehlern, bei denen die Reihenfolge
       * der zugehoerigen Cores der Reihenfolge der Cores im
       * urspruenglichen CoreInfoVector entspricht.*/
      initScheduler(remappedPcs);
   }

     
   /**
    * Fuehrt die Runde durch, indem es in einer Schleife den Scheduler
    * Zyklen durchfuehren laesst. Am Ende einer Runde versendet es die
    * Nachricht END_OF_ROUND an Tournament. Liefert false zurueck,
    * falls die Runde durch einen Aufruf von stop() beendet wurde,
    * ansonsten true.
    * 
    * @param params  Referenz auf das die eingestellten Spielparameter
    *                enthaltende ParamInfo-Objekt.
    * @return        Boolescher Wert, der angibt, ob das Turnier von aussen
    *                gestoppt wurde.
    */
   public boolean runRound(ParamInfo params)
   {
      boolean stop_flag = true;

      /**Die Schleife wird so oft durchlaufen, wie in den
       * Spielparametern Zyklen eingestellt sind. Sie wird
       * beendet, wenn entweder alle Zyklen abgeschlossen sind,
       * Stop gedrueckt wurde (dann ist stop_flag false)
       * oder alle Cores tot sind (dann ist continue_flag false).*/
      for (int i = 1;
           (i <= params.getMaxCycles()) && stop_flag && continue_flag;
           i++)
      {
         stop_flag = scheduler.runCycle();
      }

      /**Ist stop_flag noch true, d.h. es wurde nicht
       * Stop gedrueckt, wird END_OF_ROUND gemeldet.*/
      if (stop_flag)    
      {
         sendData(END_OF_ROUND, null);
      }

      /**stop_flag ist noch true, wenn nicht Stop gedrckt wurde, false sonst.*/
      return stop_flag;          
   }


   /**
    * Stoppt den laufenden Zyklus, indem die Methode stop()
    * des Schedulers aufgerufen wird.
    */
   public void stop()
   {
      scheduler.stop();  
   }


   /**
    * Versendet Daten ueber den abgelaufenen Spielzyklus an den Simulator.
    * 
    * @param id    Zahl, die codiert um welche Art Information
    *              es sich bei dem folgenden Object data handelt.
    * @param data  Object, das die eigentlichen zu
    *              uebertragenden Daten enthaelt.
    */
   public void sendData(byte id, Object data)
   {
      tournament.sendData(id, data);  
   }

   
   /**
    * Setzt die Instanzvariable continue_flag auf den bergebenen
    * Wert. Es wird false uebergeben, wenn alle Cores gestorben sind,
    * damit die Simulation dann vorzeitig beendet werden kann.
    *
    * @param continue_flag  Boolescher Wert, der false ist, wenn alle
    *                       Cores tot sind.
    */
   public void set_continue(boolean continue_flag)
   {
      this.continue_flag = continue_flag;
   }


   /**
    * Initialisiert das Ram und uebergibt eine Referenz
    * darauf an die Klassenvariable ram.
    */
   private void initRam()
   {
      this.ram = new Ram(ramSize);
   }


   /**
    * Schreibt die Cores in zufaelliger Reihenfolge und an Zufallspositionen
    * in das Ram.
    * 
    * @param minDistance     Der in den Spielparametern eingestellte
    *                        Mindestabstand zwischen den Cores.
    * @param coreInfoVector  Vector, dessen Elemente CoreInfos sind,
    *                        die die Daten zu den einzelnen Cores enthalten.
    */
   private void placeCores(short minDistance, Vector coreInfoVector)
   {
      /**Groesse des Arrays reloc, das in den CoreInfos
       * zu einem Core geliefert wird. Es enthaelt die
       * Positionen, an denen im Quelltext Labels standen,
       * deren Werte jetzt relativ zur Position des Cores
       * im Ram umgerechnet werden muessen.*/
      int reloc_length;                                           

      /**Neuen Vector anlegen, in den die geclonten Elemente
       * des coreInfoVectors geschrieben werden knnen.*/
      Vector coreInfoVectorCopy = new Vector(coreInfoVector.size());

      /**Kopie des originalen coreInfoVector anlegen,
       * um bei Manipulationen an der Kopie nicht auch
       * das Original zu veraendern.*/
      for (int i = 0; i < coreInfoVector.size(); i++)
      {
         coreInfoVectorCopy.addElement(((CoreInfo) coreInfoVector.elementAt(i)).createCopy());
      }

      /**Die Elemente aus coreInfoVector zufaellig in 
       * mappedCoreInfoVector anordnen, um auch die
       * Reihenfolge der Cores im Ram per Zufall zu waehlen
       * und nicht nur ihre Position.*/
      randomizeCoreInfoVector(coreInfoVectorCopy);     

      /**PCs zufaellig erzeugen unter Beruecksichtigung
       * von minDistance.*/
      makePcs(minDistance);                       

      /**Die folgende Schleife fuer jeden Core im coreInfoVector ausfuehren.*/
      for (int i = 0; i < mappedCoreInfoVector.size(); i++)     
      {
         /**Die folgende Schleife schreibt fr jeden Core dessen Befehle ab
          * Position PC fuer den Core in das Ram.*/
         for(int byteNumber = 0;
             byteNumber <
             (((CoreInfo) mappedCoreInfoVector.elementAt(i)).core).length;
             byteNumber++)
         {   
            /**Byte fuer Byte des jeweiligen Cores in das Ram schreiben,
             * beginnend ab der Position des PCs fr den Core.*/
            ram.write8((pcs[i] + byteNumber),
               ((CoreInfo) mappedCoreInfoVector.elementAt(i)).core[byteNumber]);    
         }
                  
         /**Feststellen, ob ein Array reloc zu dem Core existiert,
          * in dem die wegen Flags im Sourcecode anzupassenden
          * Befehle relativ zum Core-Anfang stehen.*/
         try
         {
            reloc_length =
               (((CoreInfo) mappedCoreInfoVector.elementAt(i)).reloc).length;
         }
         catch (NullPointerException e)
         {
            reloc_length = 0;
         }

         /**Die Schleife wird ausgefuehrt, falls das Array reloc existiert
          * und mindestens einen Eintrag hat.*/

         /**Die in das Ram geschriebenen Argumente der Cores korrigieren,
          * wo notwendig (falls im Sourcecode Label verwandt).*/
         for(int reloc_number = 0;
                 reloc_number < reloc_length; reloc_number = reloc_number +2)      
         {  
            /**Stelle im Core bezogen auf das erste Byte des Cores mit
             * zu aenderndem Argument.*/
            int address_to_reloc =
               ((CoreInfo) mappedCoreInfoVector.elementAt(i)).reloc[reloc_number];   

            /**Neues Argument fr die Stelle reloc aus dem Array holen und um
             * pc erhoehen, um die Relativposition zum Ram-beginn zu realisieren.*/
            int new_argument = ((CoreInfo)
               mappedCoreInfoVector.elementAt(i)).reloc[reloc_number + 1] + pcs[i];    

            while (new_argument < 0)
            {
               /**Falls der errechnete Wert negativ ist, also kleiner
                * als der PC des Core ist, wird dies durch ein "Wert" + ramSize
                * realisiert, da das Ram kreisfoermig ist.*/
               new_argument = new_argument + ramSize;            
            }

            /**Bits 0 - 7 von new_argument werden in das LSB des Arguments
             * geschrieben.*/
            ram.write8((pcs[i] + address_to_reloc), new_argument);                 

            /**Bits 8 - 15 von new_argument werden in das MSB des
             * Arguments geschrieben.*/
            ram.write8((pcs[i] + address_to_reloc + 1), (new_argument >>> 8));     
         }
      }
   }


   /**
    * Ordnet die Cores im coreInfoVector zufaellig um.
    * 
    * @param coreInfoVectorCopy  Clone des CoreInfoVector.
    */
   private void randomizeCoreInfoVector(Vector coreInfoVectorCopy)       
   {
      int size = coreInfoVectorCopy.size();
      
      /**neuen Vector mit der Groesse von coreInfoVector erzeugen.*/
      mappedCoreInfoVector = new Vector(size);             

      for (int i = 0; i < size; i++)
      {
         /**Den Namen der Cores im Vector durch ihre Position ersetzen
          * zwecks eindeutiger Identifizierung des Cores.*/
         ((CoreInfo) coreInfoVectorCopy.elementAt(i)).coreName =
            Integer.toString(i);    
      }

      for (int i = 0; i < size - 1; i++)
      {
         /**Erzeugt eine Zufallszahl zwischen 0.0 und 1.0,
          * multipliziert diese mit der Anzahl von Elementen -1
          * (da die Positionen von 0 bis anz-1 reichen)
          * in coreInfoVectorCopy und wandelt sie um in einen Integer.*/
         int random =
            Math.round((float) (Math.random() * (coreInfoVectorCopy.size() - 1))); 

         /**Das zufaellig ausgewaehlte Element aus coreInfoVectorCopy
          * an mappedCoreInfoVector anhaengen ...*/
         mappedCoreInfoVector.addElement(coreInfoVectorCopy.elementAt(random));                     

         /**... und aus coreInfoVectorCopy loeschen.*/
         coreInfoVectorCopy.removeElementAt(random);                                         
      }

      /**Wenn nur noch ein Element in coreInfoVectorCopy uebrig ist,
       * auch dieses an mappedCoreInfoVector anhaengen.*/
      mappedCoreInfoVector.addElement(coreInfoVectorCopy.elementAt(0));                             

      /**coreInfoVectorCopy aufloesen.*/
      coreInfoVectorCopy = null;                                                             
   }


   /**
    * Erzeugt zufaellige Positionen fuer die Anfaenge
    * der Cores im Ram unter Wahrung des in den
    * Spielparametern eingestellten Mindestabstandes.
    * 
    * @param minDistance  Der in den Spielparametern eingestellte
    *                     Mindestabstand zwischen den Cores.
    */
   private void makePcs(short minDistance)
   {
      /**Array mit den Gren der Cores im CoreInfoVector.*/
      short[] sizes = new short[mappedCoreInfoVector.size()];     

      /**Array mit den Abstaenden der Cores voneinander bzw.,
       * im Falle des ersten Cores, Abstand von Speicherstelle 0.*/
      int[] distances = new int[mappedCoreInfoVector.size()];     

      /**Array mit Zufallszahlen.*/
      double[] randoms = new double [mappedCoreInfoVector.size()];     

      /**Summe aller Coregroessen.*/
      int totalSize = 0;                                          

      /**Summe aller Zufallszahlen.*/
      double sum_of_randoms = 0;                                  

      /**Groesse des freien Ram nach Abzug von totalSize
       * und unter Beruecksichtigung der minDistance.*/
      int freeRam;                                                

      /**Initialisierung von pcs: einen pro Core.*/
      pcs = new int[mappedCoreInfoVector.size()];                 

      /**Die folgende Schleife wird fuer jeden Core einmal durchlaufen.*/
      for (int i = 0; i < mappedCoreInfoVector.size(); i++)
      {
         /**Groesse des Cores mit Index i im Vector in sizes[i] schreiben.*/
         sizes[i] = ((CoreInfo) mappedCoreInfoVector.elementAt(i)).size;    

         /**Summe aller Coregroessen.*/
         totalSize = totalSize + sizes[i];                            

         /**Erzeugt eine Zufallszahl zwischen 0.0 und 1.0.*/
         randoms[i] = Math.random();                                 

         /**Summe aller Zufallszahlen.*/
         sum_of_randoms = sum_of_randoms + randoms[i];              
      }

      /**Groesse des freien Ram nach Abzug von totalSize und
       * unter Beruecksichtigung der minDistance.*/
      freeRam =
         ramSize - (totalSize + (mappedCoreInfoVector.size() * minDistance));     

      for (int i = 0; i < mappedCoreInfoVector.size(); i++)
      {
         /**Coreabstaende >= minDistance werden berechnet.
          * Da durch sum_of_randoms geteilt wird, kann die Summe
          * der Abstaende nie groesser als freeRam werden.*/
         distances[i] =
            (int) (((randoms[i] * freeRam) / sum_of_randoms) + minDistance);    

         /**Speichern des PCs fuer den ersten Core, da der
          * Berechnungsmodus anders ist als fuer die naechsten Cores.*/
         if (i == 0)
         {
            pcs[i] = distances[i];                                
         }
         
         /**Speichert die mit Hilfe der Zufallszahlen erzeugten
          * PCs der Cores in einem int-Array.*/
         else
         { 
            pcs[i] = pcs[i-1] + sizes[i-1] + distances[i];        
         }
      }
   }


   /**
    * Waehlt die Reihenfolge der PCs wieder so, wie die
    * Reihenfolge der zugehoerigen Cores im urspruenglichen
    * CoreInfoVector war.
    * 
    * @return  Array von PCs, deren Reihenfolge, bezogen auf
    *          die zugehoerigen Cores, der Reihenfolge der
    *          Cores im urspruenglichen CoreInfoVector entspricht.
    */
   private int[] remapPcs()
   {
      int remappedPcs[] = new int[pcs.length];
    
      for (int i = 0; i < pcs.length; i++)
      {
         /**Die PCs wieder in der Reihenfolge in remappedPcs schreiben,
          * wie sie zu den Cores im urspruenglichen coreInfoVector gehoeren,
          * um sie an die Clients verschicken zu koennen.*/
         remappedPcs[Integer.parseInt(((CoreInfo)
            (mappedCoreInfoVector.elementAt(i))).coreName)] = pcs[i];    
      }

      return remappedPcs;
   }


   /**
    * Erzeugt ein Scheduler-Objekt an Hand eines Arrays
    * von Programmzaehlern, bei denen die Reihenfolge
    * der zugehoerigen Cores der Reihenfolge der Cores im
    * urspruenglichen CoreInfoVector entspricht.
    * 
    * @param remappedPcs  Array von PCs, deren Reihenfolge, bezogen auf
    *                     die zugehoerigen Cores, der Reihenfolge der
    *                     Cores im urspruenglichen CoreInfoVector entspricht.
    */
   private void initScheduler(int[] remappedPcs)
   {
      this.scheduler = new Scheduler(ram, remappedPcs, this);
   }

}

